# Copyright (c) 2012-2024, Abilian SAS

# ruff: noqa: RUF022
from __future__ import annotations

from .models import (
    ADMIN,
    ANONYMOUS,
    AUTHENTICATED,
    CREATE,
    CREATOR,
    DELETE,
    MANAGE,
    MANAGER,
    OWNER,
    READ,
    READER,
    WRITE,
    WRITER,
    InheritSecurity,
    Permission,
    PermissionAssignment,
    Role,
    RoleAssignment,
    RoleType,
    SecurityAudit,
)
from .service import SecurityService, security

__all__ = [
    # Permissions
    "CREATE",
    "DELETE",
    "MANAGE",
    "READ",
    "WRITE",
    # Roles
    "ADMIN",
    "ANONYMOUS",
    "AUTHENTICATED",
    "CREATOR",
    "MANAGER",
    "OWNER",
    "READER",
    "WRITER",
    #
    "InheritSecurity",
    "Permission",
    "PermissionAssignment",
    "Role",
    "RoleAssignment",
    "RoleType",
    "SecurityAudit",
    "SecurityService",
    # The security service instance
    "security",
]



# Copyright (c) 2012-2024, Abilian SAS

"""Lightweight integration and denormalisation using events (signals)."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from blinker import ANY

from abilian.core.signals import activity
from abilian.sbe.apps.documents.models import Document

from .models import Community

if TYPE_CHECKING:
    from abilian.core.entities import Entity
    from abilian.core.models.subjects import User


@activity.connect_via(ANY)
def update_community(
    sender: Any, verb: str, actor: User, object: Entity, target: Entity | None = None
) -> None:
    if isinstance(object, Community):
        object.touch()
        return

    if isinstance(target, Community):
        community = target
        community.touch()

        if isinstance(object, Document):
            if verb == "post":
                community.document_count += 1
            elif verb == "delete":
                community.document_count -= 1



# Copyright (c) 2012-2024, Abilian SAS

"""Configuration and injectable fixtures for Pytest.

Reuses fixtures defined in abilian-core.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pytest import fixture

from abilian.core.models.subjects import User
from abilian.sbe.app import create_app
from abilian.sbe.apps.communities.models import READER, Community

if TYPE_CHECKING:
    from collections.abc import Iterator

    from flask import Flask
    from flask.ctx import AppContext, RequestContext
    from flask.testing import FlaskClient
    from sqlalchemy.orm import Session

    from abilian.core.sqlalchemy import SQLAlchemy


class TestConfig:
    TESTING = True
    DEBUG = True
    SECRET_KEY = "SECRET"  # noqa: S105
    SERVER_NAME = "localhost.localdomain"
    MAIL_SENDER = "tester@example.com"
    SITE_NAME = "Abilian Test"
    PREFERRED_URL_SCHEME = "http"
    # WTF_CSRF_ENABLED = True
    WTF_CSRF_ENABLED = False
    BABEL_ACCEPT_LANGUAGES = ["en", "fr"]
    SQLALCHEMY_DATABASE_URI = "sqlite:///:memory:"


@fixture(scope="module")
def config() -> Any:
    return TestConfig


@fixture(scope="module")
def app(config: Any) -> Flask:
    # We currently return a fresh app for each test.
    # Using session-scoped app doesn't currently work.
    # Note: the impact on speed is minimal.
    # from abilian.sbe.app import create_app

    return create_app(config=config)


@fixture
def db(app: Flask, app_context: AppContext) -> Iterator[SQLAlchemy]:
    """Return a fresh db for each test."""
    from abilian.core.extensions import db as _db

    from .util import cleanup_db, ensure_services_started, stop_all_services

    stop_all_services(app)
    ensure_services_started(["blob_store", "session_blob_store"])

    cleanup_db(_db)
    _db.create_all()
    yield _db

    _db.session.remove()
    cleanup_db(_db)
    stop_all_services(app)


@fixture
def app_context(app: Flask) -> Iterator[AppContext]:
    with app.app_context() as ctx:
        yield ctx


@fixture
def test_request_context(app: Flask) -> Iterator[RequestContext]:
    with app.test_request_context() as ctx:
        yield ctx


@fixture
def session(db: SQLAlchemy) -> Session:
    return db.session


@fixture
def db_session(db: SQLAlchemy) -> Session:
    return db.session


@fixture
def user(db: SQLAlchemy) -> User:
    from abilian.core.models.subjects import User

    user = User(
        first_name="Joe",
        last_name="Test",
        email="test@example.com",
        password="test",  # noqa: S106
        can_login=True,
    )
    db.session.add(user)
    db.session.flush()
    return user


@fixture
def admin_user(db: SQLAlchemy) -> User:
    from abilian.core.models.subjects import User

    user = User(
        first_name="Jim",
        last_name="Admin",
        email="admin@example.com",
        password="admin",  # noqa: S106
        can_login=True,
    )
    user.is_admin = True
    db.session.add(user)
    db.session.flush()
    return user


@fixture
def login_user(user: User, client: FlaskClient) -> User:
    with client.session_transaction() as session:
        session["_user_id"] = user.id

    return user


@fixture
def login_admin(admin_user: User, client: FlaskClient) -> User:
    with client.session_transaction() as session:
        session["_user_id"] = admin_user.id

    return admin_user


@fixture
def community(db):
    community = Community(name="My Community")
    db.session.add(community)
    db.session.flush()
    return community


@fixture
def community1(db: SQLAlchemy) -> Community:
    community = Community(name="My Community")
    db.session.add(community)

    user = User(
        email="user_1@example.com",
        password="azerty",  # noqa: S106
        can_login=True,
    )
    db.session.add(user)
    community.set_membership(user, READER)
    community.test_user = user

    db.session.flush()
    return community


@fixture
def community2(db: SQLAlchemy) -> Community:
    community = Community(name="Another Community")
    db.session.add(community)

    user = User(
        email="user_2@example.com",
        password="azerty",  # noqa: S106
        can_login=True,
    )
    db.session.add(user)
    community.set_membership(user, READER)
    community.test_user = user

    db.session.flush()
    return community



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

from typing import TYPE_CHECKING

import jinja2

from abilian.services.preferences import preferences

from .panels.sbe_notifications import SbeNotificationsPanel

if TYPE_CHECKING:
    from abilian.app import Application


def register_plugin(app: Application) -> None:
    loader = jinja2.PackageLoader("abilian.sbe.apps.preferences")
    app.register_jinja_loaders(loader)
    preferences.register_panel(SbeNotificationsPanel(), app)



#!/usr/bin/env python

# Copyright (c) 2012-2024, Abilian SAS

"""Read public domain (CC-0) database of IP/country and export to MMDB format.

Resulting file file is written in curent working directory.
"""

from __future__ import annotations

# import csv
# import ssl
# import urllib.request
# from collections import defaultdict
# from pathlib import Path
# from tempfile import TemporaryDirectory

# import certifi
# import maxminddb
# from mmdb_writer import MMDBWriter
# from netaddr import IPRange, IPSet

CSV_DB = (
    "https://cdn.jsdelivr.net/npm/@ip-location-db/geo-whois-asn-country/"
    "geo-whois-asn-country-ipv4.csv"
)
MMDB_FILENAME = "ip_country.mmdb"


# def fetch_csv_db(csv_db_path: Path) -> None:
#     with urllib.request.urlopen(
#         CSV_DB, context=ssl.create_default_context(cafile=certifi.where())
#     ) as response:
#         csv_db_path.write_bytes(response.read())
#     size = csv_db_path.stat().st_size / 2**20
#     print(f"CSV DB: {csv_db_path.name} {size:.1f} MB")


# def write_db(db_path: Path, csv_db_path: Path):
#     data = defaultdict(list)
#     with open(csv_db_path) as file:
#         reader = csv.reader(file)
#         for line in reader:
#             data[line[2]].append(IPRange(line[0], line[1]))

#     writer = MMDBWriter()
#     for country, cidrs in data.items():
#         writer.insert_network(IPSet(cidrs), {"country": country})
#     writer.to_db_file(str(db_path))


# def check_db_done(db_path: Path) -> None:
#     free_fr_ip = "212.27.48.10"
#     ibm_com_ip = "104.85.45.20"
#     db = maxminddb.open_database(db_path)
#     result1 = db.get(free_fr_ip)
#     assert result1["country"] == "FR"
#     result2 = db.get(ibm_com_ip)
#     assert result2["country"] == "US"


# def main() -> None:
#     local_path = Path.cwd()
#     db_path = local_path / MMDB_FILENAME
#     with TemporaryDirectory(prefix="ipcountry-", dir=local_path) as temp_dir:
#         print(f"temp dir: {temp_dir}")
#         csv_db_path = Path(temp_dir) / "ip_country.csv"
#         fetch_csv_db(csv_db_path)
#         write_db(db_path, csv_db_path)
#     check_db_done(db_path)
#     size = db_path.stat().st_size // 2**20
#     print(f"MMDB: {db_path.name} {size:.1f} MB")


def main() -> None:
    print("MMDB writer not available for this release")


if __name__ == "__main__":
    main()



# Copyright (c) 2012-2024, Abilian SAS

"""Create all standard extensions."""

# Note: Because of issues with circular dependencies, Abilian-specific
# extensions are created later.
from __future__ import annotations

from . import upstream_info
from .csrf import abilian_csrf, wtf_csrf as csrf
from .jinjaext import DeferredJS
from .login import login_manager
from .mail import mail
from .redis import Redis
from .sqlalchemy import db

__all__ = (
    "csrf",
    "db",
    "login_manager",
    "mail",
    "redis",
    "upstream_info",
)

redis = Redis()

deferred_js = DeferredJS()



# Copyright (c) 2012-2024, Abilian SAS

"""Reusable widgets to be included in views.

NOTE: code is currently quite messy. Needs to be refactored.
"""

from __future__ import annotations

import base64
import html
import re
from collections import namedtuple
from datetime import datetime
from typing import TYPE_CHECKING, Any
from urllib import parse

import bleach
import sqlalchemy as sa
import sqlalchemy.orm
import wtforms
from flask import current_app, g, json, render_template, render_template_string
from flask_babel import format_date, format_datetime, format_number, get_locale
from flask_login import current_user
from flask_wtf.file import FileField
from markupsafe import Markup
from wtforms import Field, FieldList, Form, FormField, IntegerField, Label, StringField
from wtforms.widgets import (
    Input,
    PasswordInput as BasePasswordInput,
    Select,
    TextArea as BaseTextArea,
    html_params,
)
from wtforms_alchemy import ModelFieldList

from abilian.core.entities import Entity
from abilian.core.models.blob import Blob
from abilian.i18n import _, _l
from abilian.services import image
from abilian.services.image import get_format
from abilian.web import url_for
from abilian.web.filters import babel2datepicker, labelize

from .util import babel2datetime

if TYPE_CHECKING:
    from collections.abc import Sequence

    from sqlalchemy.orm.mapper import Mapper

__all__ = [
    "AjaxMainTableView",
    # Views
    "BaseTableView",
    # Widgets
    "BooleanWidget",
    "Chosen",
    "Column",
    "DateInput",
    "DateTimeInput",
    "DateTimeWidget",
    "DateWidget",
    "DefaultViewWidget",
    "EmailWidget",
    "EntityWidget",
    "FileInput",
    "FloatWidget",
    "ImageInput",
    "ListWidget",
    "MainTableView",
    "ModelListWidget",
    "MoneyWidget",
    "Panel",
    "RelatedTableView",
    "RichTextWidget",
    "Row",
    "Select2",
    "Select2Ajax",
    "SingleView",
    "TabularFieldListWidget",
    "TagInput",
    "TextArea",
    "URLWidget",
    "linkify_url",
    "text2html",
]


def linkify_url(value: str) -> str:
    """Tranform an URL pulled from the database to a safe HTML fragment."""

    value = value.strip()

    rjs = r"[\s]*(&#x.{1,7})?".join(list("javascript:"))
    rvb = r"[\s]*(&#x.{1,7})?".join(list("vbscript:"))
    re_scripts = re.compile(f"({rjs})|({rvb})", re.IGNORECASE)

    value = re_scripts.sub("", value)

    url = value
    if not url.startswith("http://") and not url.startswith("https://"):
        url = f"http://{url}"

    url = parse.urlsplit(url).geturl()
    if '"' in url:
        url = url.split('"')[0]
    if "<" in url:
        url = url.split("<")[0]

    if value.startswith("http://"):
        value = value[len("http://") :]
    elif value.startswith("https://"):
        value = value[len("https://") :]

    if value.count("/") == 1 and value.endswith("/"):
        value = value[0:-1]

    return f'<a href="{url}">{value}</a>&nbsp;<i class="fa fa-external-link"></i>'


def text2html(text: str) -> Markup | str:
    text = text.strip()
    if re.search(r"<(p|br)>", text.lower()):
        return text
    if "\n" not in text:
        return text

    lines = text.split("\n")
    lines = [line for line in lines if line]
    paragraphs = [f"<p>{line}</p>" for line in lines]
    return Markup(bleach.clean("\n".join(paragraphs), tags={"p"}))


class Column:
    def __init__(self, **kw) -> None:
        for k, w in kw.items():
            setattr(self, k, w)


class View:
    options: dict[str, Any] = {}

    def get_templates(
        self, specific_template_name: str, default_template: str
    ) -> list[str]:
        specific_template = self.options.get(specific_template_name)
        if specific_template:
            return [specific_template, default_template]
        return [default_template]


# TODO: rewrite
class BaseTableView(View):
    show_controls = False
    show_search = None
    paginate = False

    columns: list[dict[str, Any]]

    def __init__(
        self, columns: list[dict[str, Any]], options: Any | None = None
    ) -> None:
        if self.show_search is None:
            self.show_search = self.show_controls

        self.columns = []
        self.init_columns(columns)
        self.name = f"{self.__class__.__name__.lower()}-{next(g.id_generator):d}"
        if options is not None:
            self.options = options
            self.show_controls = self.options.get("show_controls", self.show_controls)
            self.show_search = self.options.get("show_search", self.show_controls)
            self.paginate = self.options.get("paginate", self.paginate)

    def init_columns(self, columns: Sequence[str | dict[str, Any]]) -> None:
        # TODO
        default_width = f"{(0.99 / len(columns) * 100):2.0f}%"
        for col in columns:
            if isinstance(col, str):
                column = {"name": col, "width": default_width}
            else:
                column = col
            assert isinstance(column, dict)
            column.setdefault("width", default_width)
            column.setdefault("sorting", ("asc", "desc"))
            if "label" not in column:
                column["label"] = labelize(column["name"])
            self.columns.append(column)

    def render(self, entities: list[Entity], **kwargs: Any) -> Markup:
        aoColumns = []
        aaSorting = []
        offset = 0
        if self.show_controls:
            aoColumns.append({"asSorting": []})
            offset = 1

        for idx, c in enumerate(self.columns, offset):
            aoColumns.append({"asSorting": c["sorting"], "sWidth": str(c["width"])})
            aaSorting.append([idx, c["sorting"][0]])

        datatable_options = {
            "aaSorting": aaSorting,
            "aoColumns": aoColumns,
            "bFilter": self.show_search,
            "oLanguage": {
                "sSearch": self.options.get("search_label", _("Filter records:"))
            },
            "bStateSave": False,
            "bPaginate": self.paginate,
            "sPaginationType": "bootstrap",
            "bLengthChange": False,
            "iDisplayLength": self.options.get("paginate_length", 50),
        }
        js = render_template_string(
            """
            require(
                ['jquery', 'jquery.dataTables'],
                function($) {
                    $('#{{ table_id  }}').dataTable({{ options|tojson|safe }});
                });
            """,
            table_id=self.name,
            options=datatable_options,
        )

        table = []
        for entity in entities:
            table.append(self.render_line(entity))

        templates = self.get_templates("template", "widgets/render_table.html")
        return Markup(
            render_template(templates, table=table, js=Markup(js), view=self, **kwargs)
        )

    def render_line(self, entity: Entity) -> list[Markup | str]:
        line = []
        make_link_on = self.options.get("make_link_on")

        for col in self.columns:
            if isinstance(col, str):
                column_name = col
                build_url = url_for
            else:
                column_name = col["name"]
                build_url = col.get("url", url_for)

            value = entity
            attr = column_name.rsplit(".", 1)
            if len(attr) > 1:
                try:
                    value = getattr(value, attr[0], "")
                    value = value.display_value(attr[1])
                except AttributeError:
                    value = ""
            else:
                value = value.display_value(attr[0])

            format = col.get("format")
            if format:
                cell = format(value)
            elif column_name in (make_link_on, "name") or col.get("linkable"):
                url = build_url(entity)
                text = html.escape(str(value))
                cell = Markup(f'<a href="{url}">{text}</a>')
            elif isinstance(value, Entity):
                url = build_url(entity)
                text = html.escape(value.name)
                cell = Markup(f'<a href="{url}">{text}</a>')
            elif isinstance(value, str) and (value.startswith(("http://", "www."))):
                cell = Markup(linkify_url(value))
            elif value in (True, False):
                cell = "\u2713" if value else ""  # Unicode "Check mark"
            elif isinstance(value, list):
                cell = "; ".join(value)
            else:
                if not isinstance(value, (Markup, str)):
                    if value is None:
                        value = ""
                    else:
                        value = str(value)
                cell = value

            line.append(cell)
        return line


class MainTableView(BaseTableView):
    """Table view for main objects list."""

    show_controls = True
    paginate = True


class RelatedTableView(BaseTableView):
    """Table view for related objects list."""

    show_controls = False
    paginate = False


class AjaxMainTableView(View):
    """Variant of the MainTableView that gets content from AJAX requests.

    TODO: refactor all of this (currently code is copy/pasted!).
    """

    show_controls = False
    paginate = True
    columns: list[dict[str, Any]]
    options: dict[str, Any]

    def __init__(
        self,
        columns: list[dict[str, Any]],
        ajax_source: str,
        search_criterions: tuple = (),
        name: str | None = None,
        options: dict[str, Any] | None = None,
    ) -> None:
        self.columns = []
        self.init_columns(columns)
        self.ajax_source = ajax_source
        self.search_criterions = search_criterions
        self.name = name if name is not None else id(self)
        self.save_state = name is not None
        self.options = options or {}

    def init_columns(self, columns: Sequence[str | dict[str, Any]]) -> None:
        # TODO: compute the correct width for each column.
        default_width = 0.99 / len(columns)
        for col in columns:
            if isinstance(col, str):
                column = {"name": col, "width": default_width}
            else:
                column = col
            assert isinstance(column, dict)
            if "label" not in column:
                name = str(column["name"])
                column["label"] = labelize(name)

            column.setdefault("sorting", ["asc", "desc"])

            if not column["sorting"]:
                column.setdefault("sortable", False)
            else:
                column.setdefault("sortable", True)

            self.columns.append(column)

    def render(self) -> Markup:
        aoColumns = [{"asSorting": []}] if self.show_controls else []
        aoColumns += [
            {"asSorting": col["sorting"], "bSortable": col["sortable"]}
            for col in self.columns
        ]
        datatable_options = {
            "sDom": "fFriltip",
            "aoColumns": aoColumns,
            "bFilter": True,
            "oLanguage": {
                "sSearch": self.options.get("search_label", _("Filter records:")),
                "oPaginate": {"sPrevious": _("Previous"), "sNext": _("Next")},
                "sLengthMenu": _("Entries per page: _MENU_"),
                "sInfo": _("Showing _START_ to _END_ of _TOTAL_ entries"),
                "sInfoEmpty": _("Showing _START_ to _END_ of _TOTAL_ entries"),
                "sInfoFiltered": _("(filtered from _MAX_ total entries)"),
                "sAddAdvancedFilter": _("Add a filter"),
                "sZeroRecords": _("No matching records found"),
                "sEmptyTable": _("No matching records found"),
            },
            "bPaginate": self.paginate,
            "sPaginationType": "bootstrap",
            "bLengthChange": True,
            "iDisplayLength": 25,
            "bStateSave": self.save_state,
            "bProcessing": True,
            "bServerSide": True,
            "sAjaxSource": self.ajax_source,
        }
        if self.options.get("aaSorting", None):
            datatable_options["aaSorting"] = self.options.get("aaSorting")

        advanced_search_filters = []
        for criterion in self.search_criterions:
            if not criterion.has_form_filter:
                continue
            d = {
                "name": criterion.name,
                "label": str(criterion.label),
                "type": criterion.form_filter_type,
                "args": criterion.form_filter_args,
                "unset": criterion.form_unset_value,
            }
            if criterion.has_form_default_value:
                d["defaultValue"] = criterion.form_default_value

            advanced_search_filters.append(d)

        if advanced_search_filters:
            datatable_options["aoAdvancedSearchFilters"] = advanced_search_filters

        return Markup(
            render_template(
                "widgets/render_ajax_table.html",
                datatable_options=datatable_options,
                view=self,
            )
        )

    def render_line(self, entity):
        line = []
        for col in self.columns:
            if isinstance(col, str):
                column_name = col
            else:
                column_name = col["name"]

            value = entity.display_value(column_name)
            has_custom_display = False

            # Manual massage.
            # 'display_fun' gets value *and* entity: useful to perform
            # specific markup based on other entity values.
            # 'display_fmt' is for simple value formatting (format_date from
            # babel for example)
            if value is None:
                value = ""
            elif "display_fun" in col:
                has_custom_display = True
                value = col["display_fun"](entity, value)
            elif "display_fmt" in col:
                value = col["display_fmt"](value)

            if has_custom_display:
                cell = value
            elif column_name == "name":
                cell = Markup(f'<a href="{url_for(entity)}">{html.escape(value)}</a>')
            elif isinstance(value, Entity):
                cell = Markup(
                    f'<a href="{url_for(value)}">{html.escape(value.name)}</a>'
                )
            elif isinstance(value, str) and (value.startswith(("http://", "www."))):
                cell = Markup(linkify_url(value))
            elif col.get("linkable"):
                cell = Markup(
                    f'<a href="{url_for(entity)}">{html.escape(str(value))}</a>'
                )
            else:
                cell = str(value)

            line.append(cell)
        return line


#
# Single object view
#
class SingleView(View):
    """View on a single object."""

    def __init__(self, form: Form, *panels: Panel, **options: Any) -> None:
        self.form = form
        self.panels = panels
        self.options = options

    def render(self, item: Entity, related_views: tuple[()] = ()) -> Markup:
        mapper = sa.orm.class_mapper(item.__class__)
        panels = []
        _to_skip = (None, False, 0, 0.0, "", "-")

        for panel in self.panels:
            data = {}
            field_name_iter = (fn for row in panel.rows for fn in row)

            for name in field_name_iter:
                field = self.form._fields[name]
                if field.is_hidden:
                    continue

                value = field.data
                if not isinstance(field, FileField) and not field.flags.render_empty:
                    if value in _to_skip:
                        continue

                value = Markup(field.render_view(entity=item))
                if value == "":
                    # related models may have [] as value, but we don't discard
                    # this type # of value in order to let widget a chance
                    # to render something useful like an 'add model' button.
                    #
                    # if it renders an empty string, there's really no point
                    # in rendering a line for this empty field
                    continue

                label = self.label_for(field, mapper, name)
                data[name] = (label, value)

            if data:
                panels.append((panel, data))

        templates = self.get_templates("view_template", "widgets/render_single.html")
        ctx = {
            "view": self,
            "related_views": related_views,
            "entity": item,
            "panels": panels,
            "form": self.form,
        }
        return Markup(render_template(templates, **ctx))

    def render_form(self, for_new=False, has_save_and_add_new=False):
        # Client-side rules for jQuery.validate
        # See: http://docs.jquery.com/Plugins/Validation/Methods#Validator
        rules = {}
        for field in self.form:
            rules_for_field = {}
            for validator in field.validators:
                rule_for_validator = getattr(validator, "rule", None)
                if rule_for_validator:
                    rules_for_field.update(rule_for_validator)
            if rules_for_field:
                rules[field.name] = rules_for_field
        if rules:
            rules = Markup(json.dumps(rules))
        else:
            rules = None

        templates = self.get_templates("edit_template", "widgets/render_for_edit.html")
        ctx = {
            "view": self,
            "form": self.form,
            "for_new": for_new,
            "has_save_and_add_new": has_save_and_add_new,
            "rules": rules,
        }
        return Markup(render_template(templates, **ctx))

    def label_for(self, field: Field, mapper: Mapper, name: str) -> Label:
        label = field.label
        if label is None:
            try:
                info = mapper.c[name].info
                label = info["label"]
            except (AttributeError, KeyError):
                pass

        if label is None:
            try:
                label = _(name)
            except KeyError:
                # i18n may be not initialized (in some unit tests for example)
                label = name

        return label


#
# Used to describe single entity views.
#
class Panel:
    """`Panel` and `Row` classes help implement a trivial internal DSL for
    specifying multi-column layouts in forms or object views.

    They are currently not really used, since we went with 1-column
    designs eventually.
    """

    def __init__(self, label: str, *rows: Row) -> None:
        self.label = label
        self.rows = rows

    def __iter__(self):
        return iter(self.rows)

    def __getitem__(self, item):
        return self.rows[item]

    def __len__(self) -> int:
        return len(self.rows)


class Row:
    """`Panel` and `Row` classes help implement a trivial internal DSL for
    specifying multi-column layouts in forms or object views.

    They are currently not really used, since we went with 1-column
    designs eventually.
    """

    def __init__(self, *cols: str) -> None:
        self.cols = cols

    def __iter__(self):
        return iter(self.cols)

    def __getitem__(self, item):
        return self.cols[item]

    def __len__(self) -> int:
        return len(self.cols)


class ModelWidget:
    edit_template = "widgets/model_widget_edit.html"
    view_template = "widgets/model_widget_view.html"

    def __init__(self, edit_template=None, view_template=None) -> None:
        if edit_template is not None:
            self.edit_template = edit_template
        if view_template is not None:
            self.view_template = view_template

    def __call__(self, field, *args, **kwargs):
        return render_template(self.edit_template, form=field)

    def render_view(self, field, *args, **kwargs):
        _to_skip = (None, False, 0, 0.0, "", "-")
        rows = []
        for f in field.form:
            if f.is_hidden:
                continue

            value = f.data
            if value in _to_skip and not f.flags.render_empty:
                continue

            value = Markup(f.render_view())
            if value == "":
                # related models may have [] as value, but we don't discard this type
                # of value in order to let widget a chance to render something useful
                # like an 'add model' button.
                #
                # if it renders an empty string, there's really no point in rendering
                # a line for this empty field
                continue

            label = f.label
            rows.append((label, value))

        return render_template(self.view_template, field=field, rows=rows)


# Form field widgets #####################################################
class TextInput(wtforms.widgets.TextInput):
    """Support pre and post icons.

    An Icon can be a plain string, or an instance of
    :class:`abilian.web.action.Icon`.
    """

    pre_icon: str | None = None
    post_icon: str | None = None

    def __init__(
        self,
        input_type: Any | None = None,
        pre_icon: Any | None = None,
        post_icon: Any | None = None,
    ) -> None:
        super().__init__(input_type)

        if pre_icon is not None:
            self.pre_icon = pre_icon
        if post_icon is not None:
            self.post_icon = post_icon

    def __call__(self, field, **kwargs):
        if not any((self.pre_icon, self.post_icon)):
            return super().__call__(field, **kwargs)

        kwargs.setdefault("type", self.input_type)
        if "value" not in kwargs:
            kwargs["value"] = field._value()

        params = self.html_params(name=field.name, **kwargs)
        return Markup(
            render_template_string(
                """
                <div class="input-group input-group-type-{{ widget.typename }}">
                {%- if widget.pre_icon %}
                  <div class="input-group-addon">{{ widget.pre_icon }}</div>
                {%- endif %}
                  <input {{ params | safe}}>
                {%- if widget.post_icon %}
                  <div class="input-group-addon">{{ widget.post_icon }}</div>
                {%- endif %}
                </div>
                """,
                widget=self,
                params=params,
            )
        )

    @property
    def typename(self):
        return self.__class__.__name__.lower()


class TextArea(BaseTextArea):
    """Accepts "resizeable" parameter: "vertical", "horizontal", "both",
    None."""

    _resizeable_valid = ("vertical", "horizontal", "both", None)
    resizeable = None
    rows = None

    def __init__(
        self,
        resizeable: str | None = None,
        rows: int | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

        if resizeable not in self._resizeable_valid:
            msg = (
                f"Invalid value for resizeable: {resizeable}, "
                f"valid values are: {self._resizeable_valid!r}"
            )
            raise ValueError(msg)
        if resizeable:
            self.resizeable = f"resizeable-{resizeable}"
        else:
            self.resizeable = "not-resizeable"

        if rows:
            self.rows = int(rows)

    def __call__(self, *args, **kwargs):
        if self.resizeable:
            css = kwargs.get("class_", "")
            kwargs["class_"] = f"{css} {self.resizeable}"

        if self.rows and "rows" not in kwargs:
            kwargs["rows"] = self.rows

        return super().__call__(*args, **kwargs)


class FileInput:
    """Renders a file input.

    Inspired from http://www.abeautifulsite.net/blog/2013/08/whipping-
    file-inputs-into-shape-with-bootstrap-3/
    """

    def __init__(self, template: str = "widgets/file_input.html") -> None:
        self.template = template

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        kwargs["name"] = field.name
        kwargs["type"] = "file"
        kwargs["disabled"] = "disabled"  # JS widget will activate it
        input_elem = f"<input {html_params(**kwargs)}>"
        button_label = _("Add file") if "multiple" in kwargs else _("Select file")

        existing = self.build_exisiting_files_list(field)
        uploads = self.build_uploads_list(field)

        if not field.multiple and uploads:
            # single file field: exising file replaced by new upload, don't show
            # existing
            existing = []
        else:
            for data in existing:
                # this might be a very big bytes object: in any case it will not be used
                # in widget's template, and during development it makes very large pages
                # due to debugtoolbar capturing template parameters
                del data["file"]

        ctx = {
            "id": field.id,
            "field": field,
            "widget": self,
            "input": input_elem,
            "button_label": button_label,
            "existing": existing,
            "uploaded": uploads,
        }
        return Markup(render_template(self.template, **ctx))

    def build_exisiting_files_list(self, field):
        existing = []
        blob = None
        object_data = field.object_data

        if not object_data:
            return existing

        if not field.multiple:
            object_data = [object_data]
            blob = field.blob

        for idx, data in enumerate(object_data):
            if data is not None:
                if hasattr(data, "content_length"):
                    size = data.content_length
                else:
                    try:
                        size = len(bytes(data))
                    except Exception:
                        # FIXME: when does it happen ?
                        size = 0
                existing.append(
                    {
                        "file": data,
                        "size": size,
                        "delete": idx in field.delete_files_index,
                    }
                )
                if hasattr(data, "filename"):
                    existing[-1]["filename"] = data.filename
                elif blob and isinstance(blob, Blob):
                    existing[-1]["filename"] = blob.meta.get("filename", "")

        return existing

    def build_uploads_list(self, field):
        uploads = current_app.extensions["uploads"]
        uploaded = []

        for handle in field.upload_handles:
            file_ = uploads.get_file(current_user, handle)
            if file_ is None:
                continue
            meta = uploads.get_metadata(current_user, handle)
            uploaded.append(
                {
                    "file": file_,
                    "handle": handle,
                    "filename": meta.get("filename", handle),
                    "size": file_.stat().st_size,
                }
            )

        return uploaded


class ImageInput(FileInput):
    """An image widget with client-side preview.

    To show current image field data has to provide an attribute named
    `url`.
    """

    def __init__(
        self,
        template: str = "widgets/image_input.html",
        width: int = 120,
        height: int = 120,
        resize_mode: str = image.CROP,
        valid_extensions: tuple[str, str, str] = ("jpg", "jpeg", "png"),
    ) -> None:
        super().__init__(template=template)
        self.resize_mode = resize_mode
        self.valid_extensions = valid_extensions
        self.width, self.height = width, height

    def build_exisiting_files_list(self, field):
        existing = super().build_exisiting_files_list(field)
        for data in existing:
            value = data["file"]
            if value:
                if hasattr(value, "url"):
                    image_url = value.url
                else:
                    image_url = self.get_b64_thumb_url(
                        self.get_thumb(value, self.width, self.height)
                    )

                data["image_url"] = image_url

        return existing

    def build_uploads_list(self, field):
        uploaded = super().build_uploads_list(field)
        for data in uploaded:
            value = data["file"]
            if value:
                if hasattr(value, "url"):
                    image_url = value.url
                else:
                    with value.open("rb") as in_:
                        image_url = self.get_b64_thumb_url(
                            self.get_thumb(in_, self.width, self.height)
                        )

                data["image_url"] = image_url

        return uploaded

    def get_thumb(self, data, width, height):
        try:
            get_format(data)
        except OSError:
            return ""
        return image.resize(data, width, height, mode=self.resize_mode)

    def get_b64_thumb_url(self, img) -> str:
        if not img:
            return ""
        try:
            fmt = image.get_format(img).lower()
        except OSError:
            return ""
        thumb = base64.b64encode(img).decode("ascii")
        return f"data:image/{fmt};base64,{thumb}"

    def render_view(self, field, **kwargs):
        data = field.data
        if not data:
            return ""
        try:
            get_format(data)
        except OSError:
            return ""

        width = kwargs.get("width", self.width)
        height = kwargs.get("heigth", self.height)

        thumb = self.get_thumb(data, width, height)
        width, height = image.get_size(thumb)

        tmpl = '<img src="{{ url }}" width="{{ width }}" height="{{ height }}" />'
        ctx = {"url": self.get_b64_thumb_url(thumb), "width": width, "height": height}
        return render_template_string(tmpl, **ctx)


class Chosen(Select):
    """Extends the Select widget using the Chosen jQuery plugin."""

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        params = html_params(name=field.name, **kwargs)
        html = [f'<select {params} class="chzn-select">']
        for val, label, selected in field.iter_choices():
            html.append(self.render_option(val, label, selected))
        html.append("</select>")
        return Markup("".join(html))

    @classmethod
    def render_option(cls, value, label, selected, **kwargs):
        options = dict(kwargs, value=value)
        if selected:
            options["selected"] = True
        params = html_params(**options)
        return Markup(f"<option {params}>{html.escape(str(label))}</option>")


class TagInput(Input):
    """Extends the Select widget using the Chosen jQuery plugin."""

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        kwargs["class_"] = "tagbox"
        if "value" not in kwargs:
            kwargs["value"] = field._value()

        params = self.html_params(name=field.name, **kwargs)
        return Markup(f"<input {params}>")


class DateInput(Input):
    """Renders date inputs using the fancy Bootstrap Datepicker:

    https://github.com/eternicode/bootstrap-datepicker
    """

    input_type = "date"

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        field_id = kwargs.pop("id")
        kwargs.setdefault("name", field.name)
        field_name = kwargs.pop("name")
        kwargs.pop("type", None)
        value = kwargs.pop("value", None)
        if value is None:
            value = field._value()
        if not value:
            value = ""

        date_fmt = kwargs.pop("format", None)
        if date_fmt is not None:
            date_fmt = (
                date_fmt.replace("%", "")
                .replace("d", "dd")
                .replace("m", "mm")
                .replace("Y", "yyyy")
            )
        else:
            date_fmt = get_locale().date_formats["short"].pattern
            date_fmt = babel2datepicker(date_fmt)
            date_fmt = date_fmt.replace("M", "m")  # force numerical months

        attributes = {
            "class": "input-group date",
            "data-provide": "datepicker",
            "data-date": value,
            "data-date-format": date_fmt,
            "data-date-autoclose": "true",
        }

        s = f"<div {html_params(**attributes)}>\n"

        s += (
            f'  <input size="13" type="text" '
            f"{html_params(name=field_name, id=field_id, value=value, **kwargs)} />\n"
        )
        s += '  <span class="input-group-addon"><i class="fa fa-calendar"></i></span>\n'
        s += "</div>\n"
        return Markup(s)

    def render_view(self, field, **kwargs):
        return format_date(field.data) if field.data else ""


class TimeInput(Input):
    """Renders time inputs using boostrap timepicker:

    https://github.com/jdewit/bootstrap-timepicker
    """

    template = "widgets/timepicker.html"

    def __init__(
        self,
        template: None = None,
        widget_mode: str = "dropdown",
        h24_mode: bool = True,
        minuteStep: int = 1,
        showSeconds: bool = False,
        secondStep: int = 1,
        showInputs: bool = False,
        disableFocus: bool = False,
        modalBackdrop: bool = False,
    ) -> None:
        super().__init__()

        if template is not None:
            self.template = template

        self.widget_mode = widget_mode
        self.h24_mode = h24_mode
        self.minuteStep = minuteStep
        self.showSeconds = showSeconds
        self.secondStep = secondStep
        self.showInputs = showInputs
        self.disableFocus = disableFocus
        self.modalBackdrop = modalBackdrop
        self.strptime = "%H:%M" if self.h24_mode else "%I:%M %p"

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        field_id = kwargs.pop("id")
        value = kwargs.pop("value", None)
        if value is None:
            value = field._value()
        if not value:
            value = ""

        time_fmt = get_locale().time_formats["short"].format
        is_h12 = "%(h)s" in time_fmt or "%(K)s" in time_fmt

        input_params = {
            "data-template": self.widget_mode,
            "data-show-meridian": is_h12,
            "data-minute-step": self.minuteStep,
            "data-show-seconds": self.showSeconds,
            "data-second-step": self.secondStep,
            "data-show-inputs": self.showInputs,
            "data-disable-focus": self.disableFocus,
            "data-modal-backdrop": self.modalBackdrop,
        }

        input_params = {k: Markup(json.dumps(v)) for k, v in input_params.items()}

        ctx = {
            "id": field_id,
            "value": value,
            "field": field,
            "required": False,
            "timepicker_attributes": input_params,
        }
        return Markup(render_template(self.template, **ctx))


class DateTimeInput:
    """if corresponding field.raw_data exist it is used to initialize default
    date & time (raw_data example: ["10/10/16 | 09:00"])"""

    def __init__(self) -> None:
        self.date = DateInput()
        self.time = TimeInput()

    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        field_id = kwargs.pop("id")
        kwargs.setdefault("name", field.name)
        field_name = kwargs.pop("name")

        locale = get_locale()
        date_fmt = locale.date_formats["short"].pattern
        date_fmt = babel2datetime(date_fmt)
        # force numerical months
        date_fmt = date_fmt.replace("%B", "%m").replace("%b", "%m")
        time_fmt = "%H:%M"

        value = kwargs.pop("value", None)
        if value is None:
            if field.data:
                value = field.data
                date_value = value.strftime(date_fmt) if value else ""
                time_value = value.strftime(time_fmt) if value else ""
            elif field.raw_data:
                value = field.raw_data  # example "10/10/16 | 09:00"
                value = "".join(value)
                value = value.split("|")

                try:
                    date_value = datetime.strptime(value[0].strip(), date_fmt)
                except (ValueError, IndexError):
                    date_value = ""

                try:
                    time_value = value[1].strip()  # example  "09:00"
                except IndexError:
                    time_value = ""

            else:
                date_value = ""
                time_value = ""

        return (
            Markup(
                '<div class="form-inline">\n'
                f'<input class="datetimepicker" type="hidden" id="{field_id}" '
                f'name="{field_name}" value="{date_value} | {time_value}" />\n'
                ""
            )
            + self.date(
                field,
                id=f"{field_id}-date",
                name=f"{field_name}-date",
                value=date_value,
            )
            + self.time(
                field,
                id=f"{field_id}-time",
                name=f"{field_name}-time",
                value=time_value,
            )
            + Markup("</div>")
        )


class DefaultViewWidget:
    def render_view(self, field: IntegerField | StringField, **kwargs: Any) -> str:
        value = field.object_data
        if isinstance(value, str):
            return text2html(value)
        # [], None and other must be rendered using empty string
        return str(value or "")


class BooleanWidget(wtforms.widgets.CheckboxInput):
    # valid data-* options when using boostrap-switch
    _ON_OFF_VALID_OPTIONS = frozenset(
        (
            "animate",
            "indeterminate",
            "inverse",
            "radio-all-off",
            "on-color",
            "off-color",
            "on-text",
            "off-text",
            "label-text",
            "handle-width",
            "label-width",
            "base-class",
            "wrapper-class",
        )
    )
    on_off_options: dict[str, Any]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.on_off_mode = kwargs.pop("on_off_mode", False)
        self.on_off_options = {}
        on_off_options = kwargs.pop("on_off_options", {})
        for k, v in on_off_options.items():
            if k not in self._ON_OFF_VALID_OPTIONS:
                continue
            self.on_off_options[f"data-{k}"] = v

        if self.on_off_mode:
            self.on_off_options["data-toggle"] = "on-off"

        super().__init__(*args, **kwargs)

    def __call__(self, field, **kwargs):
        if self.on_off_mode:
            kwargs.update(self.on_off_options)

        return super().__call__(field, **kwargs)

    def render_view(self, field, **kwargs) -> str:
        return "\u2713" if field.object_data else ""  # Text_type "Check mark"


class PasswordInput(BasePasswordInput):
    """Supports setting 'autocomplete' at instanciation time."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.autocomplete = kwargs.pop("autocomplete", None)
        super().__init__(*args, **kwargs)

    def __call__(self, field, **kwargs):
        kwargs.setdefault("autocomplete", self.autocomplete)
        return super().__call__(field, **kwargs)

    def render_view(self, field, **kwargs) -> str:
        return "*****"


class FloatWidget(wtforms.widgets.TextInput):
    """In view mode, format float number to 'precision' decimal."""

    def __init__(self, precision=None) -> None:
        self.precision = precision
        if precision is not None:
            self._fmt = f".{precision:d}f"

    def render_view(self, field, **kwargs):
        data = field.object_data
        if data is None:
            return ""

        return format(data, self._fmt)


class DateWidget(wtforms.widgets.TextInput):
    def render_view(self, field, **kwargs):
        return format_date(field.object_data) if field.object_data else ""


class DateTimeWidget(DateWidget):
    def render_view(self, field, **kwargs):
        return format_datetime(field.object_data) if field.object_data else ""


class EntityWidget:
    def render_view(self, field, **kwargs):
        objs = field.object_data
        if not field.multiple:
            objs = [objs]
        return ", ".join(
            f'<a href="{url_for(o)}">{html.escape(o.name)}</a>' for o in objs if o
        )


class HoursWidget(TextInput):
    """Widget used to show / enter hours.

    Currently hardcoded to heure(s)
    """

    post_icon = _l("hour(s)")
    input_type = "number"

    def render_view(self, field, **kwargs) -> str:
        val = field.object_data
        unit = "h"

        if val is None:
            return ""

        # \u00A0: non-breakable whitespace
        return f"{val}\u00a0{unit}"


class MoneyWidget(TextInput):
    """Widget used to show / enter money amount.

    Currently hardcoded to € / k€.
    """

    post_icon = "€"
    input_type = "number"

    def render_view(self, field, **kwargs) -> str:
        val = field.object_data
        unit = "€"

        if val is None:
            return ""

        if val > 1000:
            unit = "k€"
            val = round(val / 1000.0)

        # `format_currency()` is not used since it display numbers with cents
        # units, which we don't want
        #
        # \u00A0: non-breakable whitespace
        return f"{format_number(val)}\u00a0{unit}"


class EmailWidget(TextInput):
    pre_icon = "@"

    def render_view(self, field: StringField, **kwargs: Any) -> str:
        links = ""
        if isinstance(field, FieldList):
            for entry in field.entries:
                link = bleach.linkify(entry.data, parse_email=True)
                if link:
                    links += f' {link}&nbsp;<i class="fa fa-envelope"></i><br>'
        else:
            link = bleach.linkify(field.object_data, parse_email=True)
            if link:
                links = f'{link}&nbsp;<i class="fa fa-envelope"></i>'
        return links


class URLWidget:
    def render_view(self, field, **kwargs):
        return linkify_url(field.object_data) if field.object_data else ""


class RichTextWidget:
    template = "widgets/richtext.html"
    allowed_tags = {
        "a": {"href": True, "title": True},
        "abbr": {"title": True},
        "acronym": {"title": True},
        "b": True,
        "blockquote": True,
        "br": True,
        "code": True,
        "em": True,
        "h1": True,
        "h2": True,
        "h3": True,
        "h4": True,
        "h5": True,
        "h6": True,
        "i": True,
        "img": {"src": True},
        "li": True,
        "ol": True,
        "strong": True,
        "ul": True,
        "p": True,
        "u": True,
    }
    profile = "standard"

    def __init__(self, allowed_tags=None, template=None, profile=None) -> None:
        if allowed_tags is not None:
            self.allowed_tags = allowed_tags
        if template is not None:
            self.template = template
        if profile is not None:
            self.profile = profile

    def __call__(self, field, **kwargs):
        value = kwargs.pop("value") if "value" in kwargs else field._value()
        kwargs.setdefault("allowed_tags", self.allowed_tags)
        kwargs.setdefault("profile", self.profile)
        return render_template(self.template, field=field, value=value, kw=kwargs)


class ListWidget(wtforms.widgets.ListWidget):
    """display field label is optionnal."""

    def __init__(
        self, html_tag: str = "ul", prefix_label: bool = True, show_label: bool = True
    ) -> None:
        super().__init__(html_tag, prefix_label)
        self.show_label = show_label

    def __call__(self, field, **kwargs):
        if self.show_label:
            return super().__call__(field, **kwargs)

        kwargs.setdefault("id", field.id)
        html = [f"<{self.html_tag} {wtforms.widgets.html_params(**kwargs)}>"]
        for subfield in field:
            html.append(f"<li>{subfield()}</li>")

        html.append(f"</{self.html_tag}>")
        return Markup("".join(html))

    def render_view(self, field, **kwargs):
        data = field.data
        is_empty = data == [] if field.multiple else data is None

        if not is_empty:
            data = (
                [label for v, label, checked in field.iter_choices() if checked]
                if (hasattr(field, "iter_choices") and callable(field.iter_choices))
                else field.object_data
            )
        else:
            data = []

        tpl = """{%- for obj in data %}
              <span class="badge">{{ obj }}</span>
              {%- endfor %}"""
        return Markup(render_template_string(tpl, data=data))


class FieldListWidget:
    """For list of Fields (using <tr><td>)"""

    view_template = "widgets/fieldlist_view.html"
    template = "widgets/fieldlist.html"

    def __init__(self, template=None, view_template=None) -> None:
        if template is not None:
            self.template = template
        if view_template is not None:
            self.view_template = view_template

    def __call__(self, field, **kwargs):
        assert isinstance(field, FieldList)
        return Markup(render_template(self.template, field=field))

    def render_view(self, field, **kwargs):
        assert isinstance(field, FieldList)
        value = field.object_data
        if not value:
            return ""
        return Markup(render_template(self.view_template, field=field))


class TabularFieldListWidget:
    """For list of formfields.

    2 templates are available:

    * widgets/tabular_fieldlist_widget.html (default):
      Show sub-forms as a table, one row of inputs per model
    * widgets/model_fieldlist.html:
      Show sub-forms as a list of forms
    """

    def __init__(self, template="widgets/tabular_fieldlist_widget.html") -> None:
        self.template = template

    def __call__(self, field, **kwargs):
        assert isinstance(field, FieldList)
        labels = None

        if len(field):
            assert isinstance(field[0], FormField)
            field_names = [f.short_name for f in field[0] if not f.is_hidden]
            data_type = f"{field.entries[0].__class__.__name__}Data"
            Data = namedtuple(data_type, field_names)
            labels = Data(*(f.label for f in field[0] if not f.is_hidden))

        return Markup(render_template(self.template, labels=labels, field=field))


class ModelListWidget:
    def __init__(self, template="widgets/horizontal_table.html") -> None:
        self.template = template

    def render_view(self, field, **kwargs):
        assert isinstance(field, ModelFieldList)
        value = field.object_data
        if not value:
            return render_template(
                self.template, field=field, labels=(), rows=(), **kwargs
            )

        field_names = field._field_names
        labels = field._field_labels
        data_type = f"{field.entries[0].object_data.__class__.__name__}Data"
        Data = namedtuple(data_type, field_names)
        labels = Data(*labels)

        rows = []
        for entry in field.entries:
            row = []
            for f in filter(lambda ff: not ff.is_hidden, entry.form):
                row.append(Markup(f.render_view()))

            rows.append(Data(*row))

        rendered = render_template(
            self.template, field=field, labels=labels, rows=rows, **kwargs
        )
        return rendered


#
# Selection widget.
#
class Select2(Select):
    """Transforms a Select widget into a Select2 widget.

    Depends on global JS code.
    """

    unescape_html = False

    def __init__(
        self,
        unescape_html: bool = False,
        js_init: str = "select2",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.js_init = js_init

        if unescape_html:
            self.unescape_html = True

    def __call__(self, field, **kwargs):
        # 'placeholder' option presence is required for 'allowClear'
        params = {"placeholder": ""}
        if self.unescape_html:
            params["makeHtml"] = True
        if not field.flags.required:
            params["allowClear"] = True

        css_class = kwargs.setdefault("class_", "")
        if "js-widget" not in css_class:
            css_class += " js-widget"
            kwargs["class_"] = css_class

        kwargs.setdefault("data-init-with", self.js_init)
        kwargs["data-init-params"] = json.dumps(params)
        return super().__call__(field, **kwargs)

    def render_view(self, field, **kwargs):
        labels = [str(label) for v, label, checked in field.iter_choices() if checked]
        return "; ".join(labels)

    @classmethod
    def render_option(cls, value, label, selected, **kwargs):
        if value is None:
            return Markup("<option></option>")
        return Select.render_option(value, label, selected, **kwargs)


class Select2Ajax:
    """Ad-hoc select widget based on Select2.

    The code below is probably very fragile, since it depends on the internal
    structure of a Select2 widget.

    :param format_result: `formatResult` arg of Select2. Must be a valid
    javascript reference (like `Abilian.my_format_function`)

    :param format_selection: like `format_result`, but for Select2's
    `formatSelection`

    :param values_builder: an optional function, which given a data iterable of
    values, returns a list of dict items suitable for widget initial data;
    generally needed when using custom format_result/format_selection.
    """

    def __init__(
        self,
        template="widgets/select2ajax.html",
        multiple=False,
        format_result=None,
        format_selection=None,
        values_builder=None,
    ) -> None:
        self.template = template
        self.multiple = multiple
        self.values_builder = (
            values_builder
            if callable(values_builder)
            else lambda data: [{"id": o.id, "text": o.name} for o in data if o]
        )
        self.s2_params = {"multiple": self.multiple}

        if format_result:
            self.s2_params["formatResult"] = format_result
        if format_selection:
            self.s2_params["formatSelection"] = format_selection

    def process_formdata(self, valuelist):
        """field helper: as of select2 3.x, multiple values are passed as a
        single string."""
        if self.multiple:
            valuelist = valuelist[0].split(",")
        return valuelist

    def __call__(self, field, **kwargs):
        """Render widget."""
        if self.multiple:
            kwargs["multiple"] = True

        css_class = kwargs.setdefault("class_", "")
        if "class" in kwargs:
            css_class += kwargs.pop("class", "")
            kwargs["class_"] = css_class

        if "js-widget" not in css_class:
            css_class += " js-widget"
            kwargs["class_"] = css_class

        s2_params = {}
        s2_params.update(self.s2_params)

        if field.ajax_source:
            s2_params["ajax"] = {"url": str(field.ajax_source)}

        s2_params["placeholder"] = str(field.label.text)

        if not field.flags.required:
            s2_params["allowClear"] = True

        data = field.data
        if not self.multiple:
            data = [data]

        values = self.values_builder(data)
        input_value = ",".join(str(o.id) for o in data if o)
        data_node_id = None

        json_data = {}
        if values:
            json_data["values"] = values
            data_node_id = s2_params["dataNodeId"] = f"{field.id}-json"

        kwargs["data-init-with"] = "select2ajax"
        kwargs["data-init-params"] = json.dumps(s2_params)

        extra_args = Markup(html_params(**kwargs))

        ctx = {
            "field": field,
            "name": field.name,
            "id": field.id,
            "input_value": input_value,
            "json_data": json_data,
            "required": not field.allow_blank,
            "data_node_id": data_node_id,
            "extra_args": extra_args,
        }
        return Markup(render_template(self.template, **ctx))



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import defusedxml
import sqlalchemy as sa
import sqlalchemy.exc
from flask_migrate import Migrate
from loguru import logger

import abilian.core.util
import abilian.i18n
from abilian.core import extensions
from abilian.extensions import asset_manager
from abilian.services import (
    activity_service,
    antivirus,
    audit_service,
    auth_service,
    blob_store,
    conversion_service,
    index_service,
    preferences_service,
    security_service,
    session_blob_store,
    vocabularies_service,
)
from abilian.web import csrf
from abilian.web.action import actions
from abilian.web.admin import Admin

if TYPE_CHECKING:
    from flask import Flask

defusedxml.defuse_stdlib()

db = extensions.db

# Silence those warnings for now.
warnings.simplefilter("ignore", category=sa.exc.SAWarning)


def init_extensions(app: Flask) -> None:
    """Initialize flask extensions, helpers and services."""

    extensions.redis.init_app(app)
    extensions.mail.init_app(app)
    extensions.deferred_js.init_app(app)
    extensions.upstream_info.extension.init_app(app)

    actions.init_app(app)

    # auth_service installs a `before_request` handler (actually it's
    # flask-login). We want to authenticate user ASAP, so that sentry and
    # logs can report which user encountered any error happening later,
    # in particular in a before_request handler (like csrf validator)
    auth_service.init_app(app)

    asset_manager.init_app(app)

    init_babel(app)

    init_csrf(app)

    # Flask-Migrate
    Migrate(app, db)

    # Abilian Core services
    security_service.init_app(app)
    blob_store.init_app(app)
    session_blob_store.init_app(app)
    audit_service.init_app(app)
    index_service.init_app(app)
    activity_service.init_app(app)
    preferences_service.init_app(app)
    conversion_service.init_app(app)
    vocabularies_service.init_app(app)
    antivirus.init_app(app)

    # Admin interface
    Admin().init_app(app)


def init_babel(app: Flask) -> None:
    # Babel (for i18n)
    babel = abilian.i18n.babel
    # Temporary (?) workaround
    babel.locale_selector_func = None
    babel.timezone_selector_func = None

    babel.init_app(app)

    babel.add_translations("wtforms", translations_dir="locale", domain="wtforms")
    babel.add_translations("abilian")


def init_csrf(app: Flask) -> None:
    # CSRF by default
    if app.config.get("WTF_CSRF_ENABLED"):
        # extensions.csrf IS original wtf_csrf
        # abilian_csrf is
        # from .csrf import abilian_csrf
        # from .csrf import wtf_csrf as csrf

        extensions.csrf.init_app(app)  # double initialization
        app.extensions["csrf"] = extensions.csrf
        extensions.abilian_csrf.init_app(app)

    app.register_blueprint(csrf.csrf_blueprint)


def init_sentry(app: Flask) -> None:
    """Install Sentry handler if config defines 'SENTRY_DSN'."""
    dsn = app.config.get("SENTRY_DSN")
    if not dsn:
        return

    try:
        import sentry_sdk
    except ImportError:
        logger.error(
            'SENTRY_DSN is defined in config but package "sentry-sdk" is not installed.'
        )
        return

    from sentry_sdk.integrations.flask import FlaskIntegration
    from sentry_sdk.integrations.redis import RedisIntegration
    from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration

    sentry_sdk.init(
        dsn=dsn,
        integrations=[
            FlaskIntegration(),
            SqlalchemyIntegration(),
            RedisIntegration(),
        ],
    )



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

import time
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa
from blinker import ANY
from flask import current_app
from loguru import logger
from sqlalchemy import (
    Boolean,
    DateTime,
    ForeignKey,
    Integer,
    String,
    Unicode,
    UniqueConstraint,
    and_,
)
from sqlalchemy.event import listens_for
from sqlalchemy.orm import backref, relationship
from sqlalchemy.orm.attributes import OP_APPEND, OP_REMOVE, Event
from sqlalchemy.sql.schema import Column

from abilian.core.entities import Entity
from abilian.core.extensions import db
from abilian.core.models import NOT_AUDITABLE, SEARCHABLE
from abilian.core.models.blob import Blob
from abilian.core.models.subjects import Group, User
from abilian.i18n import _l
from abilian.sbe.apps.documents.models import Folder
from abilian.sbe.apps.documents.repository import content_repository
from abilian.services.indexing import indexable_role
from abilian.services.security import (
    ADMIN,
    MANAGER,
    READ,
    READER,
    WRITE,
    WRITER,
    Permission,
    Role,
    RoleType,
    security,
)

from . import signals

if TYPE_CHECKING:
    from werkzeug.local import LocalProxy

MEMBER = Role("member", label=_l("role_member"), assignable=False)
VALID_ROLES = frozenset([READER, WRITER, MANAGER, MEMBER])


class Membership(db.Model):
    """Represents the membership of someone in a community."""

    __tablename__ = "community_membership"

    id = Column(Integer, primary_key=True)

    user_id = Column(ForeignKey("user.id"), index=True, nullable=False)
    user = relationship(
        User, lazy="joined", backref=backref("communautes_membership", lazy="select")
    )

    community_id = Column(ForeignKey("community.id"), index=True, nullable=False)
    community = relationship("Community", lazy="joined")

    role = Column(RoleType())  # should be either 'member' or 'manager'

    __table_args__ = (UniqueConstraint("user_id", "community_id"),)

    def __repr__(self) -> str:
        return (
            f"<Membership user={self.user!r}, "
            f"community={self.community!r}, "
            f"role={self.role!s}>"
        ).encode()


def community_content(cls: type) -> Any:
    """Class decorator to mark models considered as community content.

    This is required for proper indexation.
    """
    cls.is_community_content = True

    def community_slug(self: Any) -> str:
        return self.community and self.community.slug

    cls.community_slug = property(community_slug)

    index_to = getattr(cls, "__index_to__", ())
    index_to += (("community_slug", ("community_slug",)),)
    cls.__index_to__ = index_to

    def _indexable_roles_and_users(self: Any) -> str:
        if not self.community:
            return []
        return indexable_roles_and_users(self.community)

    cls._indexable_roles_and_users = property(_indexable_roles_and_users)
    return cls


def indexable_roles_and_users(community: Community) -> str:
    """Mixin to use to replace Entity._indexable_roles_and_users.

    Will be removed when communities are upgraded to use standard role
    based access (by setting permissions and using security service).
    """
    # TODO: remove
    return " ".join(indexable_role(user) for user in community.members)


class Community(Entity):
    """Ad-hoc objects that hold properties about a community."""

    __tablename__ = "community"
    __index_to__ = (("id", ("id", "community_id")),)

    is_community_content = True

    # : A public description.
    description = Column(Unicode(500), default="", nullable=False, info=SEARCHABLE)

    #: An image or logo for this community.
    image_id = Column(ForeignKey(Blob.id), index=True)
    image = relationship(Blob, lazy="joined")

    #: The root folder for this community.
    folder_id = Column(
        Integer,
        ForeignKey(Folder.id, use_alter=True, name="fk_community_root_folder"),
        unique=True,
    )
    folder = relationship(
        Folder,
        single_parent=True,  # required for delete-orphan
        primaryjoin=(folder_id == Folder.id),
        cascade="all, delete-orphan",
        backref=backref("_community", lazy="select", uselist=False),
    )

    #: The group this community is linked to, if any. Memberships are then
    #: reflected
    group_id = Column(ForeignKey(Group.id), nullable=True, unique=True)
    group = relationship(Group, foreign_keys=group_id, lazy="select")

    #: Memberships for this community.
    memberships = relationship(Membership, cascade="all, delete-orphan")

    #: direct access to :class:`User` members
    members = relationship(
        User,
        secondary=Membership.__table__,
        viewonly=True,
        backref=backref("communities", lazy="select", viewonly=True),
    )

    #: Number of members in this community.
    membership_count = Column(Integer, default=0, nullable=False, info=NOT_AUDITABLE)

    #: Number of documents in this community.
    document_count = Column(Integer, default=0, nullable=False, info=NOT_AUDITABLE)

    #: Last time something happened in this community
    last_active_at = Column(
        DateTime, default=datetime.utcnow, nullable=False, info=NOT_AUDITABLE
    )

    # Various features for this community:

    #: True if this community has a document management space
    has_documents = Column(Boolean, nullable=False, default=True)

    #: True if this community has a wiki
    has_wiki = Column(Boolean, nullable=False, default=True)

    #: True if this community has a forum
    has_forum = Column(Boolean, nullable=False, default=True)

    #: One of 'participative' or 'informative
    type = Column(String(20), nullable=False, default="informative")

    #: One of 'secret', 'public', 'open' (currently not used)
    visibility = Column(String(20), nullable=False, default="secret")

    #: Used for segmenting communities. Not used currently.
    category = Column(String(20), nullable=False, default="")

    #: True if regular members can send stuff by email to all members
    # Not used anymore.
    members_can_send_by_email = Column(Boolean, nullable=False, default=False)

    def __init__(self, **kw) -> None:
        self.has_documents = True
        self.membership_count = 0
        self.document_count = 0
        self.members_can_send_by_email = False
        super().__init__(**kw)
        if self.has_documents and not self.folder:
            # FIXME: this should be done in documents by using signals
            name = self.name
            if not name:
                # during creation, we may have to provide a temporary name for
                # subfolder, we don't want empty names on folders since they must be
                # unique among siblings
                name = f"{self.__class__.__name__}_{self.id!s}-{time.asctime()}"
            self.folder = content_repository.root_folder.create_subfolder(name)
            # if not self.group:
            #   self.group = Group(name=self.name)

        if not self.image:
            fn = Path(__file__).parent / "views" / "data" / "community.png"
            self.image = Blob(fn.open("rb").read())

    @property
    def has_calendar(self) -> bool:
        config = current_app.config
        return bool(config.get("ENABLE_CALENDAR"))

    def rename(self, name: str) -> None:
        self.name = name
        if self.folder:
            # FIXME: use signals
            self.folder.name = name

    def get_memberships(self, role: str | None = None) -> list[Membership]:
        M = Membership
        memberships = M.query.filter(M.community_id == self.id)
        if role:
            memberships = memberships.filter(M.role == role)
        return memberships.all()

    def set_membership(self, user: User, role: str | Role) -> None:
        """Add a member with the given role, or set the role of an existing
        member."""
        assert isinstance(user, User)
        role = Role(role)

        if role not in VALID_ROLES:
            msg = f"Invalid role: {role}"
            raise ValueError(msg)

        session = sa.orm.object_session(self) or db.session()
        is_new = True
        M = Membership
        membership = (
            session.query(M)
            .filter(and_(M.user_id == user.id, M.community_id == self.id))
            .first()
        )

        if not membership:
            membership = Membership(community=self, user=user, role=role)
            session.add(membership)
            self.membership_count += 1
        else:
            is_new = False
            membership.role = role

        signals.membership_set.send(self, membership=membership, is_new=is_new)

    def remove_membership(self, user: User) -> None:
        M = Membership
        membership = M.query.filter(
            and_(M.user_id == user.id, M.community_id == self.id)
        ).first()
        if not membership:
            msg = f"User {user} is not a member of community {self}"
            raise KeyError(msg)

        db.session.delete(membership)
        self.membership_count -= 1
        signals.membership_removed.send(self, membership=membership)

    def update_roles_on_folder(self) -> None:
        if self.folder:
            self.ungrant_all_roles_on_folder()
            for membership in self.memberships:
                user = membership.user
                role = membership.role
                if role == MANAGER:
                    security.grant_role(user, MANAGER, self.folder)
                elif self.type == "participative":
                    security.grant_role(user, WRITER, self.folder)
                else:
                    security.grant_role(user, READER, self.folder)

    def ungrant_all_roles_on_folder(self) -> None:
        # Only needed if the object is not being created
        if self.folder and self.id:
            role_assignments = security.get_role_assignements(self.folder)
            for principal, role in role_assignments:
                security.ungrant_role(principal, role, self.folder)

    def get_role(self, user: User | LocalProxy) -> Role | None:
        """Returns the given user's role in this community."""
        M = Membership
        membership = (
            db.session()
            .query(M.role)
            .filter(and_(M.community_id == self.id, M.user_id == user.id))
            .first()
        )

        return membership.role if membership else None

    def has_member(self, user) -> bool:
        return self.get_role(user) is not None

    def has_permission(self, user: LocalProxy, permission: Permission) -> bool:
        if not isinstance(permission, Permission):
            assert isinstance(permission, str)
            permission = Permission(permission)

        if user.has_role(ADMIN):
            return True

        role = self.get_role(user)
        return role == MANAGER or (role == MEMBER and permission in (READ, WRITE))

    def touch(self) -> None:
        self.last_active_at = datetime.utcnow()

    @property
    def _indexable_roles_and_users(self) -> str:
        return indexable_roles_and_users(self)


def CommunityIdColumn() -> Column:
    return Column(
        ForeignKey(Community.id),
        nullable=False,
        info=SEARCHABLE | {"index_to": (("community_id", ("community_id",)),)},
    )


# Handlers to keep community/group members in sync
#
_PROCESSED_ATTR = "__sbe_community_group_sync_processed__"


@signals.membership_set.connect_via(ANY)
def _membership_added(sender: Community, membership: Membership, is_new: bool) -> None:
    if not is_new:
        return

    if getattr(membership.user, _PROCESSED_ATTR, False) is OP_APPEND:
        return

    if sender.group and membership.user not in sender.group.members:
        logger.debug(
            "_membership_added({sender}, {membership}, {is_new}) user: {user}",
            sender=repr(sender),
            membership=repr(membership),
            is_new=repr(is_new),
            user=repr(membership.user),
        )
        setattr(membership.user, _PROCESSED_ATTR, OP_APPEND)
        sender.group.members.add(membership.user)


@signals.membership_removed.connect_via(ANY)
def membership_removed(sender: Community, membership: Membership) -> None:
    if getattr(membership.user, _PROCESSED_ATTR, False) is OP_REMOVE:
        return

    if sender.group and membership.user in sender.group.members:
        logger.debug(
            "_membership_removed({sender}, {membership}) user: {user}",
            sender=repr(sender),
            membership=repr(membership),
            user=repr(membership.user),
        )
        setattr(membership.user, _PROCESSED_ATTR, OP_REMOVE)
        sender.group.members.discard(membership.user)


@listens_for(Community.members, "append")
@listens_for(Community.members, "remove")
def _on_member_change(community, user, initiator) -> None:
    group = community.group
    if not group:
        return

    logger.debug(
        "_on_member_change({community}, {user}, op={initiator})",
        community=repr(community),
        user=repr(user),
        initiator=repr(initiator.op),
    )

    if getattr(user, _PROCESSED_ATTR, False) is initiator.op:
        return

    setattr(user, _PROCESSED_ATTR, initiator.op)

    if initiator.op is OP_APPEND:
        if user not in group.members:
            group.members.add(user)

    elif initiator.op is OP_REMOVE:
        if user in group.members:
            group.members.discard(user)


@listens_for(Community.group, "set", active_history=True)
def _on_linked_group_change(
    community: Community, value: Any, oldvalue: Any, initiator: Event
) -> None:
    if value == oldvalue:
        return

    logger.debug(
        "_on_linked_group_change({community}, {value}, {oldvalue})",
        community=repr(community),
        value=repr(value),
        oldvalue=repr(oldvalue),
    )

    if oldvalue is not None and oldvalue.members:
        logger.debug(
            "_on_linked_group_change({community}, {value}, {oldvalue}): oldvalue clear()",
            community=repr(community),
            value=repr(value),
            oldvalue=repr(oldvalue),
        )
        oldvalue.members.clear()

    members = set(community.members)
    if value is not None and value.members != members:
        logger.debug(
            "_on_linked_group_change({community}, {value}, {oldvalue}): set value.members",
            community=repr(community),
            value=repr(value),
            oldvalue=repr(oldvalue),
        )
        value.members = members


def _safe_get_community(group: Group) -> Community | None:
    session = sa.orm.object_session(group)
    if not session:
        return None

    with session.no_autoflush:
        try:
            return (
                session.query(Community)
                .filter(Community.group == group)
                .options(
                    sa.orm.joinedload(Community.group),
                    sa.orm.joinedload(Community.members),
                )
                .one()
            )
        except sa.orm.exc.NoResultFound:
            return None


@listens_for(Group.members, "append")
@listens_for(Group.members, "remove")
def _on_group_member_change(group: Group, user: User, initiator: Event) -> None:
    community = _safe_get_community(group)

    if not community:
        return

    op = initiator.op
    if getattr(user, _PROCESSED_ATTR, False) is op:
        return

    is_present = user in community.members
    setattr(user, _PROCESSED_ATTR, op)
    logger.debug(
        "_on_group_member_change({group}, {user}, op={initiator}) "
        "community: {community}",
        group=repr(group),
        user=repr(user),
        initiator=repr(initiator.op),
        community=repr(community),
    )

    if (op is OP_APPEND and is_present) or (op is OP_REMOVE and not is_present):
        return

    if op is OP_APPEND:
        community.set_membership(user, MEMBER)

    elif op is OP_REMOVE:
        community.remove_membership(user)


@listens_for(Group.members, "set", active_history=True)
def _on_group_members_replace(group, value, oldvalue, initiator) -> None:
    if value == oldvalue:
        return

    community = _safe_get_community(group)
    if not community:
        return

    members = set(community.members)
    logger.debug(
        "_on_group_members_replace({group}, {value}, {oldvalue}) "
        "community: {community}",
        group=repr(group),
        value=repr(value),
        oldvalue=repr(oldvalue),
        community=repr(community),
    )

    for u in members - value:
        if getattr(u, _PROCESSED_ATTR, False) is OP_REMOVE:
            continue
        setattr(u, _PROCESSED_ATTR, OP_REMOVE)
        community.remove_membership(u)

    for u in value - members:
        if getattr(u, _PROCESSED_ATTR, False) is OP_APPEND:
            continue
        setattr(u, _PROCESSED_ATTR, OP_APPEND)
        community.set_membership(u, MEMBER)



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

import importlib
from typing import TYPE_CHECKING

from attrs import field, frozen

if TYPE_CHECKING:
    from flask import Flask

CORE_PLUGINS = [
    "abilian.web.search",
    "abilian.web.tags",
    "abilian.web.comments",
    "abilian.web.uploads",
    "abilian.web.attachments",
]


@frozen
class PluginManager:
    """Mixin that provides support for loading plugins."""

    app: Flask
    registered_plugins: set[str] = field(factory=set)

    def register_plugins(self, plugins: list[str]) -> None:
        """Load plugins listed in config variable 'PLUGINS'."""
        for plugin_fqdn in plugins:
            if plugin_fqdn not in self.registered_plugins:
                self._register_plugin(plugin_fqdn)
                self.registered_plugins.add(plugin_fqdn)

    def _register_plugin(self, name: str) -> None:
        """Load and register a plugin given its package name."""
        module = importlib.import_module(name)
        module.register_plugin(self.app)  # type: ignore



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from flask import Flask


def setup_blueprints(app: Flask) -> None:
    from abilian.services.preferences import preferences
    from abilian.web.coreviews import users
    from abilian.web.preferences.user import UserPreferencesPanel
    from abilian.web.views.images import images_bp

    app.register_blueprint(images_bp)
    app.register_blueprint(users.blueprint)
    preferences.register_panel(UserPreferencesPanel(), app)



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

import logging
import sys
from typing import TYPE_CHECKING

from loguru import logger

if TYPE_CHECKING:
    from flask import Flask


def init_logging(app: Flask) -> None:
    run_from_cli = app.config.get("RUN_FROM_CLI")

    if run_from_cli and not app.debug:
        print("Running from CLI, skipping logging init (use --debug to enable)")
        logger.configure(handlers=[])
    else:
        init_loguru(app)


class InterceptHandler(logging.Handler):
    def emit(self, record) -> None:
        logger_opt = logger.opt(depth=6, exception=record.exc_info)
        logger_opt.log(record.levelno, record.getMessage())


def init_loguru(app: Flask) -> None:
    """Register loguru as (sole) handler."""
    if app.debug:
        level = "DEBUG"
    else:
        level = "INFO"

    logger.remove()
    logger.add(sys.stderr, level=level)

    app.logger.handlers = [InterceptHandler()]
    logger.info("Loguru initialized")



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

import sqlite3
from typing import Any

import sqlalchemy as sa
import sqlalchemy.exc
import sqlalchemy.pool
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.event import listens_for


@listens_for(sa.pool.Pool, "checkout")
def ping_connection(
    dbapi_connection: Connection, connection_record, connection_proxy
) -> None:
    """Ensure connections are valid.

    From: `http://docs.sqlalchemy.org/en/rel_0_8/core/pooling.html`

    In case db has been restarted pool may return invalid connections.
    """
    cursor = dbapi_connection.cursor()
    try:
        cursor.execute("SELECT 1")
    except Exception as e:
        # optional - dispose the whole pool
        # instead of invalidating one at a time
        # connection_proxy._pool.dispose()

        # raise DisconnectionError - pool will try
        # connecting again up to three times before raising.
        raise sa.exc.DisconnectionError from e
    cursor.close()


#
# Make Sqlite a bit more well-behaved.
#
@sa.event.listens_for(Engine, "connect")
def _set_sqlite_pragma(dbapi_connection: Any, connection_record: Any) -> None:
    if isinstance(dbapi_connection, sqlite3.Connection):  # pragma: no cover
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON;")
        cursor.close()



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

from itertools import count
from typing import TYPE_CHECKING, Any

import jinja2
import sqlalchemy as sa
import sqlalchemy.orm
import svcs
from flask import Flask, appcontext_pushed, g, request_started
from flask_super import register_commands
from flask_talisman import DEFAULT_CSP_POLICY, Talisman

from abilian.core import extensions, signals
from abilian.extensions import asset_manager
from abilian.services import settings_service
from abilian.services.security import ANONYMOUS
from abilian.web.access_blueprint import allow_access_for_roles
from abilian.web.nav import setup_nav_and_breadcrumbs

from .blueprints import setup_blueprints
from .debug import setup_debug
from .extensions import init_extensions, init_sentry
from .logging import init_logging

if TYPE_CHECKING:
    from abilian.app import Application


def setup_app(app: Application, plugins=None) -> None:
    # At this point we have loaded all external config files:
    # SQLALCHEMY_DATABASE_URI is definitively fixed (it cannot be defined in
    # database AFAICT), and LOGGING_FILE cannot be set in DB settings.

    svcs.flask.init_app(app)

    init_logging(app)
    init_sentry(app)

    # time to load config bits from database: 'settings'
    # First init required stuff: db to make queries, and settings service
    extensions.db.init_app(app)
    settings_service.init_app(app)

    init_extensions(app)

    app.register_jinja_loaders(jinja2.PackageLoader("abilian.web"))

    app.install_default_handlers()

    if app.debug and not app.testing:
        setup_debug(app)

    if not app.debug:
        # CSP
        csp = app.config.get("CONTENT_SECURITY_POLICY", DEFAULT_CSP_POLICY)
        Talisman(app, content_security_policy=csp)

    setup_blueprints(app)

    plugins = (plugins or []) + list(app.config["PLUGINS"])

    with app.app_context():
        app.plugin_manager.register_plugins(plugins)

    init_access_controllers(app)

    asset_manager.finalize_assets_setup(app)

    # At this point all models should have been imported: time to configure
    # mappers. Normally Sqlalchemy does it when needed but mappers may be
    # configured inside sa.orm.class_mapper() which hides a
    # misconfiguration: if a mapper is misconfigured its exception is
    # swallowed by class_mapper(model) results in this laconic
    # (and misleading) message: "model is not mapped"
    sa.orm.configure_mappers()

    if not app.testing:
        with app.app_context():
            # Time to generate the JS API
            signals.register_js_api.send(app)

            # Initialize Abilian core services.
            # Must come after all entity classes have been declared.
            # Delegated to ServiceManager. Will need some configuration love
            # later.
            app.service_manager.start_services()

    connect_signals()

    # Register CLI commands
    register_commands(app)


def connect_signals() -> None:
    appcontext_pushed.connect(install_id_generator)
    request_started.connect(setup_nav_and_breadcrumbs)


# TODO: remove
def install_id_generator(sender: Flask, **kwargs: Any) -> None:
    g.id_generator = count(start=1)


def init_access_controllers(app: Application) -> None:
    app.add_access_controller(
        "static", allow_access_for_roles(ANONYMOUS), endpoint=True
    )
    # debugtoolbar: this is needed to have it when not authenticated
    # on a private site. We cannot do this in init_debug_toolbar,
    # since auth service is not yet installed.
    app.add_access_controller(
        "debugtoolbar",
        allow_access_for_roles(ANONYMOUS),
    )
    app.add_access_controller(
        "_debug_toolbar.static",
        allow_access_for_roles(ANONYMOUS),
        endpoint=True,
    )



# Copyright (c) 2012-2024, Abilian SAS

"""I18n.

To mark strings for translation::

    from abilian.i18n import _
    _(u'message to translate')

Use :data:`_` for gettext, :data:`_l` for lazy_gettext, :data:`_n` for
ngettext.

:data:`Babel extension<babel>` support multiple translation paths. This
allows to add more catalogs to search for translations, in LIFO
order. This feature can be used to override some translations in a
custom application, be providing a catalog with messages to override::

    current_app.extensions['babel'].add_translations('abilian.core')

See :meth:`add_translations<Babel.add_translations>`.

To extract messages to build the message catalog template (.pot), use the
following "`-k`" parameters:

.. code-block:: bash

    $ pybabel extract -F babel.cfg -k "_n:1,2" -k "_l" -o "msg.pot" "src"


This can be made easier by placing in `setup.cfg`:

.. code-block:: ini

    [extract_messages]
    mapping_file = babel.cfg
    keywords = _n:1,2 _l
    output-file = msg.pot
    input-dirs = src


And just type:

.. code-block:: bash

    $ python setup.py extract_messages
"""

from __future__ import annotations

import importlib
import os
import re
import unicodedata
from datetime import datetime, tzinfo
from gettext import GNUTranslations
from pathlib import Path
from typing import TYPE_CHECKING, Any

import flask_babel
import pytz
from babel import Locale
from babel.dates import LOCALTZ, get_timezone, get_timezone_gmt
from babel.localedata import locale_identifiers
from babel.support import Translations as BaseTranslations
from flask import Flask, current_app, g, render_template, request
from flask_babel import (
    Babel as BabelBase,
    LazyString,
    force_locale,
    gettext,
    lazy_gettext,
    ngettext,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

__all__ = [
    "VALID_LANGUAGES_CODE",
    "_",
    "_l",
    "_n",
    "babel",
    "get_default_locale",
    "gettext",
    "lazy_country_name",
    "lazy_gettext",
    "localeselector",
    "ngettext",
    "render_template_i18n",
    "timezoneselector",
]

#: gettext alias
_ = gettext
#: lazy_gettext alias
_l = lazy_gettext
#: ngettext alias
_n = ngettext

#: accepted languages codes
VALID_LANGUAGES_CODE = frozenset(
    _lang for _lang in locale_identifiers() if len(_lang) == 2
)


def get_default_locale() -> Locale:
    return current_app.extensions["babel"].default_locale


def _get_locale() -> Locale:
    locale = flask_babel.get_locale()
    if locale is None:
        locale = get_default_locale()
    return locale


def _gettext_territory(code):
    locale = _get_locale()
    return locale.territories.get(code) or get_default_locale().territories.get(code)


#: get localized territory name
def country_name(code):
    return _gettext_territory(code)


#: lazy version of :func:`country_name`
def lazy_country_name(code):
    return LazyString(_gettext_territory, code)


def default_country() -> str:
    return current_app.config.get("DEFAULT_COUNTRY", "")


def country_choices(
    first: str = "", default_country_first: bool = True
) -> list[tuple[str, str]]:
    """Return a list of (code, countries), alphabetically sorted on localized
    country name.

    :param first: Country code to be placed at the top
    :param default_country_first:
    :type default_country_first: bool
    """
    locale = _get_locale()
    territories = [
        (code, name) for code, name in locale.territories.items() if len(code) == 2
    ]  # skip 3-digit regions

    if first == "" and default_country_first:
        first = default_country()

    def sortkey(item: tuple[str, str]) -> str:
        if first is not None and item[0] == first:
            return "0"
        return to_lower_ascii(item[1])

    territories.sort(key=sortkey)
    return territories


def supported_app_locales() -> Iterator[tuple[Locale, str]]:
    """Language codes and labels supported by current application.

    :return: an iterable of `(:class:`babel.Locale`, label)`, label being the
    locale language human name in current locale.
    """
    locale = _get_locale()
    codes = current_app.config["BABEL_ACCEPT_LANGUAGES"]
    return ((Locale.parse(code), locale.languages.get(code, code)) for code in codes)


def timezones_choices():
    """Timezones values and their labels for current locale.

    :return: an iterable of `(code, label)`, code being a timezone code and label
    the timezone name in current locale.
    """
    utcnow = pytz.utc.localize(datetime.utcnow())
    locale = _get_locale()
    for tz in sorted(pytz.common_timezones):
        tz = get_timezone(tz)
        now = tz.normalize(utcnow.astimezone(tz))
        label = f"({get_timezone_gmt(now, locale=locale)}) {tz.zone}"
        yield (tz, label)  # get_timezone_name(tz, locale=locale))


class Babel(BabelBase):
    """Allow to load translations from other modules."""

    _translations_paths: list[tuple[str, str]]

    def init_app(
        self,
        app: Flask,
        locale_selector: Callable | None = None,
        timezone_selector: Callable | None = None,
    ) -> None:
        if locale_selector is None:
            locale_selector = localeselector
        if timezone_selector is None:
            timezone_selector = timezoneselector
        if hasattr(self, "localeselector"):
            # Old Babel version
            super().init_app(app)
            self.localeselector(locale_selector)
            self.timezoneselector(timezone_selector)
        else:
            # recent Babel version
            super().init_app(
                app,
                locale_selector=locale_selector,
                timezone_selector=timezone_selector,
            )
        assert app.root_path
        self._translations_paths = [
            (os.path.join(app.root_path, "translations"), "messages")
        ]

    def add_translations(
        self,
        module_name: str,
        translations_dir: str = "translations",
        domain: str = "messages",
    ) -> None:
        """Add translations from external module.

        For example::

            babel.add_translations('abilian.core')

        Will add translations files from `abilian.core` module.
        """
        module = importlib.import_module(module_name)
        for path in (Path(p, translations_dir) for p in module.__path__):
            if not (path.exists() and path.is_dir()):
                continue

            if not os.access(str(path), os.R_OK):
                self.app.logger.warning(
                    "Babel translations: read access not allowed {path}, skipping.",
                    path=str(path),
                )
                continue

            self._translations_paths.append((str(path), domain))


class Translations(BaseTranslations):
    """Merge only non-empty translations.

    This avoids having uncomplete catalog that "clear" existing
    translations, when used with :func:`_get_translations_multi_paths`.
    """

    def merge(self, translations: Translations) -> Translations:
        if isinstance(translations, GNUTranslations):
            for msgkey, msgstr in translations._catalog.items():
                msgid = msgkey

                if isinstance(msgkey, tuple):
                    msgid = msgkey[0]

                msgstr = msgstr.strip()
                if msgkey in self._catalog and (msgid == msgstr):
                    # when msgstr is empty, compile_catalog sets msgstr = msgid
                    # so this is probable an existing translation that would
                    # be "erased" by msgid string: skip it.

                    # logger.debug('Catalog: %r, skip msgkey: %r, existing: %r',
                    # translations, msgkey, self._catalog[msgkey])
                    continue

                self._catalog[msgkey] = msgstr

            if isinstance(translations, BaseTranslations):
                self.files.extend(translations.files)

        return self


def _get_translations_multi_paths() -> Translations | None:
    """Return the correct gettext translations that should be used for this
    request.

    This will never fail and return a dummy translation object if used
    outside of the request or if a translation cannot be found.
    """
    if not request:
        return None

    translations = getattr(g, "babel_translations", None)
    if translations is None:
        babel_ext = current_app.extensions["babel"]
        translations = None
        trs = None

        # reverse order: thus the application catalog is loaded last, so that
        # translations from libraries can be overriden
        for dirname, domain in reversed(babel_ext._translations_paths):
            trs = Translations.load(
                dirname, locales=[flask_babel.get_locale()], domain=domain
            )

            # babel.support.Translations is a subclass of
            # babel.support.NullTranslations, so we test if object has a 'merge'
            # method

            if not trs or not hasattr(trs, "merge"):
                # got None or NullTranslations instance
                continue

            if translations is not None and hasattr(translations, "merge"):
                translations.merge(trs)
            else:
                translations = trs

        # ensure translations is at least a NullTranslations object
        if translations is None:
            translations = trs

        g.babel_translations = translations

    return translations


# monkey patch flask-babel
flask_babel.get_translations = _get_translations_multi_paths

#: importable instance of :class:`Babel`
babel = Babel()


def localeselector() -> str | None:
    """Default locale selector used in abilian applications."""
    # if a user is logged in, use the locale from the user settings
    user = getattr(g, "user", None)
    if user is not None:
        locale = getattr(user, "locale", None)
        if locale:
            return locale

    # Otherwise, try to guess the language from the user accept header the browser
    # transmits.  By default we support en/fr. The best match wins.
    return request.accept_languages.best_match(
        current_app.config["BABEL_ACCEPT_LANGUAGES"]
    )


def timezoneselector() -> tzinfo:
    """Default timezone selector used in abilian applications."""
    return LOCALTZ


def get_template_i18n(template_name: str, locale: Locale) -> list[str]:
    """Build template list with preceding locale if found."""
    if locale is None:
        return [template_name]

    template_list = []
    parts = template_name.rsplit(".", 1)
    root = parts[0]
    suffix = parts[1]

    if locale.territory is not None:
        locale_string = f"{locale.language}_{locale.territory}"
        localized_template_path = f"{root}.{locale_string}.{suffix}"
        template_list.append(localized_template_path)

    localized_template_path = f"{root}.{locale.language}.{suffix}"
    template_list.append(localized_template_path)

    # append the default
    template_list.append(template_name)
    return template_list


class ensure_request_context:
    """Context manager that ensures a request context is set up."""

    _rq_ctx = None

    def __enter__(self):
        if not request:
            ctx = self._rq_ctx = current_app.test_request_context()
            ctx.__enter__()

    def __exit__(self, *args: object):
        ctx = self._rq_ctx
        self._rq_ctx = None

        if ctx is not None:
            ctx.__exit__(*args)


def render_template_i18n(template_name_or_list: str, **context: Any) -> str:
    """Try to build an ordered list of template to satisfy the current
    locale."""
    template_list: list[str] = []
    # Use locale if present in **context
    if "locale" in context:
        locale = Locale.parse(context["locale"])
    else:
        # Use get_locale() or default_locale
        locale = flask_babel.get_locale()

    if isinstance(template_name_or_list, str):
        template_list = get_template_i18n(template_name_or_list, locale)
    else:
        # Search for locale for each member of the list, do not bypass
        for template in template_name_or_list:
            template_list.extend(get_template_i18n(template, locale))

    with ensure_request_context(), force_locale(locale):
        return render_template(template_list, **context)


_NOT_WORD_RE = re.compile(r"[^\w\s]+", flags=re.UNICODE)


def to_lower_ascii(value: str) -> str:
    value = str(value)
    value = _NOT_WORD_RE.sub(" ", value)
    value = unicodedata.normalize("NFKD", value)
    value_b = value.encode("ascii", "ignore")
    value = value_b.decode("ascii")
    value = value.strip().lower()
    value = re.sub(r"[_\s]+", " ", value)
    return value



# Copyright (c) 2012-2024, Abilian SAS

"""Calendar module."""

from __future__ import annotations


def register_plugin(app) -> None:
    from .actions import register_actions
    from .views import blueprint

    blueprint.record_once(register_actions)
    app.register_blueprint(blueprint)



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

from flask import Blueprint, Flask, abort
from loguru import logger

from abilian.core import extensions


def setup_debug(app: Flask) -> None:
    init_debug_toolbar(app)
    init_debug_pages(app)


def init_debug_toolbar(app: Flask) -> None:
    if not app.debug or app.testing:
        return

    try:
        from flask_debugtoolbar import DebugToolbarExtension
    except ImportError:
        logger.warning("Running in debug mode but flask_debugtoolbar is not installed.")
        return

    dbt = DebugToolbarExtension()
    default_config = dbt._default_config(app)
    init_dbt = dbt.init_app

    if "DEBUG_TB_PANELS" not in app.config:
        # add our panels to default ones
        app.config["DEBUG_TB_PANELS"] = list(default_config["DEBUG_TB_PANELS"])
    init_dbt(app)
    for view_name in app.view_functions:
        if view_name.startswith("debugtoolbar."):
            extensions.csrf.exempt(app.view_functions[view_name])


def init_debug_pages(app: Flask) -> None:
    # dev helper
    # during dev, one can go to /http_error/403 to see rendering of 403
    http_error_pages = Blueprint("http_error_pages", __name__)

    @http_error_pages.route("/<int:code>")
    def error_page(code) -> None:
        """Helper for development to show 403, 404, 500..."""
        abort(code)

    app.register_blueprint(http_error_pages, url_prefix="/http_error")



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from flask import Flask


def register_plugin(app: Flask) -> None:
    from .views import wall

    app.register_blueprint(wall)



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

from typing import TYPE_CHECKING

from attrs import field, frozen

if TYPE_CHECKING:
    from abilian.services import Service


@frozen
class ServiceManager:
    """Provides lifecycle (register/start/stop) support for services."""

    services: dict[str, Service] = field(factory=dict)

    def add_service(self, name: str, service: Service) -> None:
        self.services[name] = service

    def get_service(self, name: str) -> Service:
        return self.services[name]

    def start_services(self, services: list[str] | None = None) -> None:
        """Start all services. If a service is already running, nothing happens."""
        if services is None:
            services = self.services.values()
        for service in services:
            if not service.running:
                service.start()

    def stop_services(self) -> None:
        """Stop all services. If a service is not running, nothing happens."""
        for service in self.services.values():
            if service.running:
                service.stop()

    def list_services(self):
        return self.services.values()



"""
empty message

Revision ID: 1f4631da751b
Revises: 946bada3d9d5
Create Date: 2017-05-24 20:19:54.896355
"""

# revision identifiers, used by Alembic.
from __future__ import annotations

revision = "1f4631da751b"
down_revision = "946bada3d9d5"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_foreign_key(
        "fk_inherited_entity_id", "sbe_event", "entity", ["id"], ["id"], use_alter=True
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("fk_inherited_entity_id", "sbe_event", type_="foreignkey")
    # ### end Alembic commands ###



"""
empty message

Revision ID: 946bada3d9d5
Revises:
Create Date: 2017-05-16 13:48:19.338776
"""

# revision identifiers, used by Alembic.
from __future__ import annotations

revision = "946bada3d9d5"
down_revision = None
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "sbe_event",
        sa.Column("community_id", sa.Integer(), nullable=False),
        sa.Column("title", sa.Unicode(), nullable=False),
        sa.Column("description", sa.Unicode(), nullable=False),
        sa.Column("location", sa.Unicode(), nullable=False),
        sa.Column("start", sa.DateTime(), nullable=False),
        sa.Column("end", sa.DateTime(), nullable=True),
        sa.Column("url", sa.Unicode(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["community_id"],
            ["community.id"],
        ),
        sa.ForeignKeyConstraint(
            ["id"], ["entity.id"], name="fk_inherited_entity_id", use_alter=True
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "view",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("entity_id", sa.Integer(), nullable=False),
        sa.Column("_fk_entity_id", sa.Integer(), nullable=True),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["_fk_entity_id"], ["entity.id"], ondelete="SET NULL"),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "hit",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("view_id", sa.Integer(), nullable=True),
        sa.Column("viewed_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["view_id"],
            ["view.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.drop_constraint(
        "administratorship_group_id_fkey", "administratorship", type_="foreignkey"
    )
    op.drop_constraint(
        "administratorship_user_id_fkey", "administratorship", type_="foreignkey"
    )
    op.create_foreign_key(
        None,
        "administratorship",
        "group",
        ["group_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        None,
        "administratorship",
        "user",
        ["user_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.add_column(
        "forum_thread", sa.Column("last_post_at", sa.DateTime(), nullable=True)
    )
    op.drop_constraint("membership_user_id_fkey", "membership", type_="foreignkey")
    op.drop_constraint("membership_group_id_fkey", "membership", type_="foreignkey")
    op.create_foreign_key(
        None,
        "membership",
        "user",
        ["user_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        None,
        "membership",
        "group",
        ["group_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "membership", type_="foreignkey")
    op.drop_constraint(None, "membership", type_="foreignkey")
    op.create_foreign_key(
        "membership_group_id_fkey", "membership", "group", ["group_id"], ["id"]
    )
    op.create_foreign_key(
        "membership_user_id_fkey", "membership", "user", ["user_id"], ["id"]
    )
    op.drop_column("forum_thread", "last_post_at")
    op.drop_constraint(None, "administratorship", type_="foreignkey")
    op.drop_constraint(None, "administratorship", type_="foreignkey")
    op.create_foreign_key(
        "administratorship_user_id_fkey",
        "administratorship",
        "user",
        ["user_id"],
        ["id"],
    )
    op.create_foreign_key(
        "administratorship_group_id_fkey",
        "administratorship",
        "group",
        ["group_id"],
        ["id"],
    )
    op.drop_table("hit")
    op.drop_table("view")
    op.drop_table("sbe_event")
    # ### end Alembic commands ###



# Copyright (c) 2012-2024, Abilian SAS

""""""

from __future__ import annotations

from .models import Vocabulary
from .service import get_vocabulary, vocabularies

__all__ = ["Vocabulary", "get_vocabulary", "vocabularies"]



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

from . import documents, folders
from .views import community_blueprint

__all__ = ("community_blueprint", "documents", "folders")



# Copyright (c) 2012-2024, Abilian SAS

"""Navigation elements.

Abilian define theses categories:   `section`:     Used for navigation
elements relevant to site section   `user`:     User for element that
should appear in user menu
"""

from __future__ import annotations

import typing
from typing import Any

from flask import Flask, g, request
from jinja2 import Template
from markupsafe import Markup

from .action import ACTIVE, ENABLED, Action, Glyphicon, getset

if typing.TYPE_CHECKING:
    from flask_babel.speaklater import LazyString

    from abilian.web.action import Endpoint, Status


def setup_nav_and_breadcrumbs(_app: Flask) -> None:
    """Listener for `request_started` event.

    If you want to customize first items of breadcrumbs, override
    :meth:`init_breadcrumbs`
    """
    g.nav = {"active": None}  # active section
    g.breadcrumb = []
    init_breadcrumbs()


def init_breadcrumbs() -> None:
    """Insert the first element in breadcrumbs.

    This happens during `request_started` event, which is triggered
    before any url_value_preprocessor and `before_request` handlers.
    """
    g.breadcrumb.append(BreadcrumbItem(icon="home", url=f"/{request.script_root}"))


class NavItem(Action):
    """A single navigation item."""

    divider = False

    def __init__(
        self, category: str, name: str, divider: bool = False, *args: Any, **kwargs: Any
    ) -> None:
        category = f"navigation:{category}"
        super().__init__(category, name, *args, **kwargs)
        self.divider = divider

    @getset
    def status(self, value: Any | None = None) -> Status:
        current = g.nav.get("active")
        if current is None:
            return ENABLED

        if not current.startswith("navigation:"):
            current = f"navigation:{current}"

        status = ACTIVE if current == self.path else ENABLED
        return status

    @property
    def path(self) -> str:
        return f"{self.category}:{self.name}"


class NavGroup(NavItem):
    """A navigation group renders a list of items."""

    template_string = """
    <ul class="nav navbar-nav {{ action.css_class }}">
      <li class="dropdown">
        <a class="dropdown-toggle" data-toggle="dropdown">
          {%- if action.icon %}{{ action.icon }}{% endif %}
          {{ action.title }} <b class="caret"></b>
        </a>
        <ul class="dropdown-menu">
          {%- for item in action_items %}
          {%- if item.divider %}<li class="divider"></li>{%- endif %}
          <li class="{{ item.status|safe }}">{{ item.render() }}</li>
          {%- endfor %}
        </ul>
      </li>
    </ul>
    """

    def __init__(
        self, category: str, name: str, items: tuple[NavItem] = (), *args, **kwargs
    ) -> None:
        super().__init__(category, name, *args, **kwargs)
        self.items = list(items)
        self._paths = {self.path}
        for i in self.items:
            self._paths.add(i.path)

    def append(self, item: NavItem) -> None:
        self.items.append(item)
        self._paths.add(item.path)

    def insert(self, pos: int, item: NavItem) -> None:
        self.items.insert(pos, item)
        self._paths.add(item.path)

    def get_render_args(self, **kwargs: Any) -> dict[str, Any]:
        params = super().get_render_args(**kwargs)
        params["action_items"] = [a for a in self.items if a.available(params)]
        return params

    @getset
    def status(self, value: Any | None = None) -> Status:
        current = g.nav.get("active")
        if current is None:
            return ENABLED

        if not current.startswith("navigation:"):
            current = f"navigation:{current}"
        status = ACTIVE if current in self._paths else ENABLED
        return status


class BreadcrumbItem:
    """A breadcrumb element has at least a label or an icon."""

    #: Label shown to user. May be an i18n string instance
    label = None

    #: Icon to use.
    icon = None

    #: Additional text, can be used as tooltip for example
    description = None

    #: either an Unicode string or an :class:`Endpoint` instance.
    _url = None

    template_string = (
        '{%- if url %}<a href="{{ url }}">{%- endif %}'
        "{%- if item.icon %}{{ item.icon }}\u00a0{%- endif %}"
        "{{ item.label }}"
        "{%- if url %}</a>{%- endif %}"
    )

    def __init__(
        self,
        label: LazyString | str = "",
        url: str | Endpoint = "#",
        icon: str | None = None,
        description: Any | None = None,
    ) -> None:
        # don't test 'label or...': if label is a lazy_gettext, it will be
        # resolved. If this item is created in a url_value_preprocessor, it will
        # setup i18n before auth has loaded user, so i18n will fallback on browser
        # negociation instead of user's site preference, and load wrong catalogs for
        # the whole request.
        assert label is not None or icon is None
        self.label = label
        if isinstance(icon, str):
            icon = Glyphicon(icon)

        self.icon = icon
        self.description = description
        self._url = url
        self.__template = Template(self.template_string)

    @property
    def url(self) -> str:
        return str(self._url)

    def render(self) -> Markup:
        return Markup(self.__template.render(item=self, url=self.url))



# Copyright (c) 2012-2024, Abilian SAS

"""Conversion service.

Hardcoded to manage only conversion to PDF, to text and to image series.

Includes result caching (on filesystem).

Assumes poppler-utils and LibreOffice are installed.

TODO: rename Converter into ConversionService ?
"""

from __future__ import annotations

import shutil
import subprocess
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING

from loguru import logger
from PIL import Image
from PIL.ExifTags import TAGS

from .cache import Cache
from .exceptions import HandlerNotFoundError
from .handler_lock import init_conversion_lock_dir
from .handlers import Handler, poppler_bin_util
from .util import make_temp_file

if TYPE_CHECKING:
    from flask import Flask

TMP_DIR = "tmp"
CACHE_DIR = "cache"


class Converter:
    tmp_dir: Path
    cache_dir: Path
    handlers: list[Handler]

    def __init__(self) -> None:
        self.handlers = []
        self.cache = Cache()

    def init_app(self, app: Flask) -> None:
        self.init_work_dirs(
            cache_dir=Path(app.instance_path, CACHE_DIR),
            tmp_dir=Path(app.instance_path, TMP_DIR),
        )
        init_conversion_lock_dir(app.instance_path)
        app.extensions["conversion"] = self

        for handler in self.handlers:
            handler.init_app(app)

    def init_work_dirs(self, cache_dir: Path, tmp_dir: Path) -> None:
        self.tmp_dir = tmp_dir
        self.cache_dir = cache_dir
        self.cache.cache_dir = self.cache_dir

        if not self.tmp_dir.exists():
            self.tmp_dir.mkdir()
        if not self.cache_dir.exists():
            self.cache_dir.mkdir()

    def clear(self) -> None:
        self.cache.clear()
        for d in (self.tmp_dir, self.cache_dir):
            shutil.rmtree(bytes(d))
            d.mkdir()

    def register_handler(self, handler: Handler) -> None:
        self.handlers.append(handler)

    # TODO: refactor, pass a "File" or "Document" or "Blob" object
    def to_pdf(self, digest: str, blob: bytes, mime_type: str) -> bytes:
        cache_key = ("pdf", digest)
        pdf = self.cache.get_bytes(cache_key)
        if pdf:
            return pdf
        for handler in self.handlers:
            if handler.accept(mime_type, "application/pdf"):
                pdf = handler.convert(blob)
                if pdf is None:
                    continue
                self.cache[cache_key] = pdf
                return pdf
        msg = f"No handler found to convert from {mime_type} to PDF"
        raise HandlerNotFoundError(msg)

    def to_text(self, digest: str, blob: bytes, mime_type: str) -> str:
        """Convert a file to plain text.

        Useful for full-text indexing. Returns a Unicode string.
        """
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return ""

        cache_key = ("txt", digest)

        text = self.cache.get_text(cache_key)
        if text:
            return text

        # Direct conversion possible
        for handler in self.handlers:
            if handler.accept(mime_type, "text/plain"):
                text = handler.convert(blob)
                self.cache[cache_key] = text
                return text

        # Use PDF as a pivot format
        pdf = self.to_pdf(digest, blob, mime_type)
        for handler in self.handlers:
            if handler.accept("application/pdf", "text/plain"):
                text = handler.convert(pdf)
                self.cache[cache_key] = text
                return text

        msg = f"No handler found to convert from {mime_type} to text"
        raise HandlerNotFoundError(msg)

    def has_image(self, digest, mime_type, index, size=500):
        """Tell if there is a preview image."""
        cache_key = (f"img:{index}:{size}", digest)
        return mime_type.startswith("image/") or cache_key in self.cache

    def get_image(self, digest, blob, mime_type, index, size=500):
        """Return an image for the given content, only if it already exists in
        the image cache."""
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return ""

        cache_key = (f"img:{index}:{size}", digest)
        return self.cache.get(cache_key)

    def to_image(
        self, digest: str, blob: bytes, mime_type: str, index: int, size: int = 500
    ) -> bytes:
        """Convert a file to a list of images.

        Returns image at the given index.
        """
        # Special case, for now (XXX).
        if mime_type.startswith("image/"):
            return b""

        cache_key = (f"img:{index}:{size}", digest)
        converted = self.cache.get_bytes(cache_key)
        if converted:
            return converted

        # Direct conversion possible
        for handler in self.handlers:
            if handler.accept(mime_type, "image/jpeg"):
                converted_images = handler.convert(blob, size=size)
                for i in range(len(converted_images)):
                    converted = converted_images[i]
                    cache_key = (f"img:{i}:{size}", digest)
                    self.cache[cache_key] = converted
                return converted_images[index]

        # Use PDF as a pivot format
        pdf = self.to_pdf(digest, blob, mime_type)
        for handler in self.handlers:
            if handler.accept("application/pdf", "image/jpeg"):
                converted_images = handler.convert(pdf, size=size)
                for i in range(len(converted_images)):
                    converted = converted_images[i]
                    cache_key = (f"img:{i}:{size}", digest)
                    self.cache[cache_key] = converted
                return converted_images[index]

        msg = f"No handler found to convert from {mime_type} to image"
        raise HandlerNotFoundError(msg)

    def get_metadata(self, digest, content, mime_type):
        """Get a dictionary representing the metadata embedded in the given
        content."""

        # XXX: ad-hoc for now, refactor later
        if mime_type.startswith("image/"):
            img = Image.open(BytesIO(content))
            ret = {}
            if not hasattr(img, "_getexif"):
                return {}
            info = img._getexif()
            if not info:
                return {}
            for tag, value in info.items():
                decoded = TAGS.get(tag, tag)
                ret[f"EXIF:{decoded!s}"] = value
            return ret

        if mime_type != "application/pdf":
            content = self.to_pdf(digest, content, mime_type)

        with make_temp_file(content) as in_fn:
            pdfinfo = poppler_bin_util("pdfinfo") or "pdfinfo"
            try:
                output = subprocess.check_output([pdfinfo, in_fn])
            except OSError:  # pragma: no cover
                logger.error("Conversion failed, probably pdfinfo is not installed")
                raise

        ret = {}
        for line in output.split(b"\n"):
            if b":" in line:
                key, value = line.strip().split(b":", 1)
                key = str(key)
                ret[f"PDF:{key}"] = str(value.strip(), errors="replace")

        return ret



# Copyright (c) 2012-2024, Abilian SAS

"""Create a SQLAlchemy instance and configure it."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import sqlalchemy as sa
import sqlalchemy.event
import sqlalchemy.orm

from abilian.core.sqlalchemy import SQLAlchemy

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection
    from sqlalchemy.sql.schema import MetaData

__all__ = ("db",)

db = SQLAlchemy()


@sa.event.listens_for(db.metadata, "before_create")
@sa.event.listens_for(db.metadata, "before_drop")
def _filter_metadata_for_connection(
    target: MetaData, connection: Connection, **kw: Any
) -> None:
    """Listener to control what indexes get created.

    Useful for skipping postgres-specific indexes on a sqlite for example.

    It's looking for info entry `engines` on an index
    (`Index(info=dict(engines=['postgresql']))`), an iterable of engine names.
    """
    engine = connection.engine.name
    default_engines = (engine,)
    tables = target if isinstance(target, sa.Table) else kw.get("tables", [])
    for table in tables:
        indexes = list(table.indexes)
        for idx in indexes:
            if engine not in idx.info.get("engines", default_engines):
                table.indexes.remove(idx)


def _install_get_display_value(cls: Any) -> None:
    _MARK = object()

    def display_value(self, field_name, value=_MARK):
        """Return display value for fields having 'choices' mapping (stored
        value.

        -> human-readable value). For other fields it will simply return field
        value.

        `display_value` should be used instead of directly getting field value.

        If `value` is provided, it is "translated" to a human-readable value. This is
        useful for obtaining a human-readable label from a raw value.
        """
        val = getattr(self, field_name, "ERROR") if value is _MARK else value

        mapper = sa.orm.object_mapper(self)
        try:
            field = getattr(mapper.c, field_name)
        except AttributeError:
            pass
        else:
            if "choices" in field.info:

                def get(v):
                    return field.info["choices"].get(v, v)

                if isinstance(val, list):
                    val = [get(v) for v in val]
                else:
                    val = get(val)

        return val

    if not hasattr(cls, "display_value"):
        cls.display_value = display_value


sa.event.listen(db.Model, "class_instrument", _install_get_display_value)



# Copyright (c) 2012-2024, Abilian SAS

"""Base Flask application class, used by tests or to be extended in real
applications."""

from __future__ import annotations

import errno
import os
import sys
import warnings
from functools import cached_property, partial
from pathlib import Path
from typing import TYPE_CHECKING, Any

import defusedxml
import sqlalchemy as sa
import sqlalchemy.exc
from flask import Flask
from flask.config import ConfigAttribute
from loguru import logger

import abilian.core.util
import abilian.i18n
from abilian.config import default_config
from abilian.core import extensions, signals
from abilian.core.plugin_manager import PluginManager
from abilian.core.service_manager import ServiceManager
from abilian.lib.scanner import scan_package
from abilian.services import auth_service
from abilian.services.security import ANONYMOUS
from abilian.setup import setup_app
from abilian.web.access_blueprint import allow_access_for_roles
from abilian.web.errors import ErrorManagerMixin
from abilian.web.jinja import JinjaManagerMixin
from abilian.web.util import send_file_from_directory
from abilian.web.views import Registry as ViewRegistry

if TYPE_CHECKING:
    from collections.abc import Callable, Collection

    from abilian.services.security.models import Role

__all__ = ["Application", "create_app"]

defusedxml.defuse_stdlib()
# Silence those warnings for now.
warnings.simplefilter("ignore", category=sa.exc.SAWarning)

db = extensions.db


def create_app(config: type | None = None, plugins=None, **kw: Any) -> Application:
    app = Application(**kw)

    # 1: Ensure app.config is properly set up
    app.configure(config)
    app.check_instance_folder(create=True)

    # 2: Scan to pre-register callbacks, services, etc.
    scan_package("abilian.cli")

    # 3: Perform registrations on app
    setup_app(app, plugins=plugins)

    # 4: Perform post-registration actions
    signals.components_registered.send(app)

    return app


class Application(
    # AssetManagerMixin,
    ErrorManagerMixin,
    JinjaManagerMixin,
    Flask,
):
    """Base application class.

    Extend it in your own app.
    """

    default_config = default_config

    #: If True all views will require by default an authenticated user, unless
    #: Anonymous role is authorized. Static assets are always public.
    private_site = ConfigAttribute("PRIVATE_SITE")

    #: instance of :class:`.web.views.registry.Registry`.
    default_view: ViewRegistry

    #: json serializable dict to land in Javascript under Abilian.api
    js_api: dict[str, Any]

    def __init__(self, name: Any | None = None, *args: Any, **kwargs: Any) -> None:
        name = name or __name__

        Flask.__init__(self, name, *args, **kwargs)

        self.service_manager = ServiceManager()
        self.plugin_manager = PluginManager(self)

        JinjaManagerMixin.__init__(self)

        self.default_view = ViewRegistry()
        self.js_api = {}

    @cached_property
    def data_dir(self) -> Path:
        path = Path(self.instance_path, "data")
        if not path.exists():
            path.mkdir(0o775, parents=True)

        return path

    def configure(self, config: type | None) -> None:
        if config:
            # This is usually the case for tests
            self.config.from_object(config)
        else:
            self.config.from_prefixed_env()

        # Setup babel config
        languages = self.config["BABEL_ACCEPT_LANGUAGES"]
        languages = tuple(
            lang for lang in languages if lang in abilian.i18n.VALID_LANGUAGES_CODE
        )
        self.config["BABEL_ACCEPT_LANGUAGES"] = languages

        # This needs to be done dynamically
        if not self.config.get("SESSION_COOKIE_NAME"):
            self.config["SESSION_COOKIE_NAME"] = f"{self.name}-session"

        if not self.config.get("FAVICO_URL"):
            self.config["FAVICO_URL"] = self.config.get("LOGO_URL")

        if not self.debug and self.config["SECRET_KEY"] == "CHANGEME":  # noqa: S105
            logger.error("You must change the default secret config ('SECRET_KEY')")
            sys.exit()

    def check_instance_folder(self, create=False) -> None:
        """Verify instance folder exists, is a directory, and has necessary
        permissions.

        :param:create: if `True`, creates directory hierarchy

        :raises: OSError with relevant errno if something is wrong.
        """
        path = Path(self.instance_path)
        err = None
        eno = 0

        if not path.exists():
            if create:
                logger.info("Create instance folder: {path}", path=str(path))
                path.mkdir(0o775, parents=True, exist_ok=True)
            else:
                err = "Instance folder does not exists"
                eno = errno.ENOENT
        elif not path.is_dir():
            err = "Instance folder is not a directory"
            eno = errno.ENOTDIR
        elif not os.access(path, os.R_OK | os.W_OK | os.X_OK):
            err = 'Require "rwx" access rights, please verify permissions'
            eno = errno.EPERM

        if err:
            raise OSError(eno, err, str(path))

    def add_url_rule_with_role(
        self,
        rule: str,
        endpoint: str,
        view_func: Callable,
        roles: Collection[Role] = (),
        **options: Any,
    ) -> None:
        """See :meth:`Flask.add_url_rule`.

        If `roles` parameter is present, it must be a
        :class:`abilian.service.security.models.Role` instance, or a list of
        Role instances.
        """
        self.add_url_rule(rule, endpoint, view_func, **options)

        if roles:
            self.add_access_controller(
                endpoint, allow_access_for_roles(roles), endpoint=True
            )

    def add_access_controller(
        self, name: str, func: Callable, endpoint: bool = False
    ) -> None:
        """Add an access controller.

        If `name` is None it is added at application level, else if is
        considered as a blueprint name. If `endpoint` is True then it is
        considered as an endpoint.
        """
        auth_state = self.extensions[auth_service.name]

        if endpoint:
            if not isinstance(name, str):
                msg = f"{name!r} is not a valid endpoint name"
                raise ValueError(msg)

            auth_state.add_endpoint_access_controller(name, func)
        else:
            auth_state.add_bp_access_controller(name, func)

    def add_static_url(
        self, url_path: str, directory: str, endpoint: str, roles: Collection[Role] = ()
    ) -> None:
        """Add a new url rule for static files.

        :param url_path: subpath from application static url path. No heading
            or trailing slash.
        :param directory: directory to serve content from.
        :param endpoint: flask endpoint name for this url rule.

        Example::

            app.add_static_url(
                'myplugin',
                '/path/to/myplugin/resources',
                endpoint='myplugin_static')

        With default setup it will serve content from directory
        `/path/to/myplugin/resources` from url `http://.../static/myplugin`
        """
        url_path = f"{self.static_url_path}/{url_path}/<path:filename>"
        self.add_url_rule_with_role(
            url_path,
            endpoint=endpoint,
            view_func=partial(send_file_from_directory, directory=directory),
            roles=roles,
        )
        self.add_access_controller(
            endpoint, allow_access_for_roles(ANONYMOUS), endpoint=True
        )



# Copyright (c) 2012-2024, Abilian SAS

"""Application settings service."""

from __future__ import annotations

from .models import register
from .service import SettingsService

__all__ = ["SettingsService", "register"]



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from flask import Flask, url_for
from flask_babel import lazy_gettext as _l
from flask_login import current_user

from abilian.sbe.apps.communities.actions import CommunityEndpoint
from abilian.sbe.apps.communities.security import is_manager
from abilian.services import get_service
from abilian.services.security import ADMIN, SecurityService
from abilian.web.action import Action, FAIcon, ModalActionMixin, actions

if TYPE_CHECKING:
    from flask.blueprints import BlueprintSetupState


class WikiPageAction(Action):
    Endpoint = CommunityEndpoint

    def pre_condition(self, context: dict[str, Any]) -> bool:
        page = context.get("object")
        return bool(page)

    def url(self, context=None):
        if self._url:
            return self._url
        page = context.get("object")
        kw = self.endpoint.get_kwargs()
        kw["title"] = page.title
        return url_for(self.endpoint.name, **kw)


def is_admin(context):
    security = cast("SecurityService", get_service("security"))
    return security.has_role(current_user, ADMIN, object=context.get("object"))


class WikiPageModalAction(ModalActionMixin, WikiPageAction):
    pass


class WikiAction(Action):
    Endpoint = CommunityEndpoint


_actions = (
    WikiPageAction(
        "wiki:page",
        "page_viewers",
        _l("Readers list"),
        icon="user",
        condition=lambda ctx: is_manager(context=ctx),
        endpoint=".page_viewers",
    ),
    WikiPageAction(
        "wiki:page",
        "view",
        _l("View"),
        endpoint=".page",
        icon="eye-open",
    ),
    WikiPageAction(
        "wiki:page",
        "edit",
        _l("Edit"),
        endpoint=".page_edit",
        icon="pencil",
    ),
    WikiPageModalAction(
        "wiki:page",
        "upload_attachment",
        _l("Upload an attachment"),
        url="#upload-files",
        icon="plus",
    ),
    WikiPageAction(
        "wiki:page",
        "source",
        _l("Source"),
        endpoint=".page_source",
        icon=FAIcon("code"),
    ),
    WikiPageAction(
        "wiki:page",
        "changes",
        _l("Changes"),
        endpoint=".page_changes",
        icon="time",
    ),
    WikiPageModalAction(
        "wiki:page",
        "delete",
        _l("Delete"),
        url="#modal-delete",
        icon="trash",
    ),
    WikiAction(
        "wiki:global",
        "new",
        _l("New page"),
        endpoint=".page_new",
        icon="plus",
    ),
    WikiAction(
        "wiki:global",
        "pages",
        _l("All pages"),
        endpoint=".wiki_pages",
        icon="list",
    ),
    WikiAction(
        "wiki:global",
        "help",
        _l("Syntax help"),
        endpoint=".wiki_help",
        icon="info-sign",
    ),
)


def register_actions(state: BlueprintSetupState) -> None:
    if not actions.installed(state.app):
        return

    app = cast("Flask", state.app)
    with app.app_context():
        actions.register(*_actions)



# Copyright (c) 2012-2024, Abilian SAS

"""Additional data types for sqlalchemy."""

from __future__ import annotations

import contextlib
import json
import uuid
from typing import TYPE_CHECKING, Any

import babel
import babel.dates
import pytz
import sqlalchemy as sa
import sqlalchemy.dialects
from sqlalchemy.ext.mutable import Mutable

if TYPE_CHECKING:
    from sqlalchemy.engine.interfaces import Dialect
    from sqlalchemy.sql.sqltypes import CHAR


class MutationDict(Mutable, dict):
    """Provides a dictionary type with mutability support."""

    @classmethod
    def coerce(cls, key: str, value: dict) -> MutationDict:
        """Convert plain dictionaries to MutationDict."""
        if not isinstance(value, MutationDict):
            if isinstance(value, dict):
                return MutationDict(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        return value

    #  pickling support. see:
    #  http://docs.sqlalchemy.org/en/rel_0_8/orm/extensions/mutable.html#supporting-pickling
    def __getstate__(self) -> dict:
        return dict(self)

    def __setstate__(self, state: dict):
        self.update(state)

    # dict methods
    def __setitem__(self, key: str, value: int | str) -> None:
        """Detect dictionary set events and emit change events."""
        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key) -> None:
        """Detect dictionary del events and emit change events."""
        dict.__delitem__(self, key)
        self.changed()

    def clear(self) -> None:
        dict.clear(self)
        self.changed()

    def update(self, other, **kw) -> None:
        dict.update(self, other, **kw)
        self.changed()

    def setdefault(self, key, failobj=None):
        if key not in self:
            self.changed()
        return dict.setdefault(self, key, failobj)

    def pop(self, key, *args):
        self.changed()
        return dict.pop(self, key, *args)

    def popitem(self):
        self.changed()
        return dict.popitem(self)


class MutationList(Mutable, list):
    """Provides a list type with mutability support."""

    @classmethod
    def coerce(cls, key: str, value: list) -> MutationList:
        """Convert list to MutationList."""
        if not isinstance(value, MutationList):
            if isinstance(value, list):
                return MutationList(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        return value

    #  pickling support. see:
    #  http://docs.sqlalchemy.org/en/rel_0_8/orm/extensions/mutable.html#supporting-pickling
    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop("_parents", None)
        return d

    # list methods
    def __setitem__(self, idx, value) -> None:
        list.__setitem__(self, idx, value)
        self.changed()

    def __delitem__(self, idx: Any) -> None:
        list.__delitem__(self, idx)
        self.changed()

    def insert(self, idx, value) -> None:
        list.insert(self, idx, value)
        self.changed()

    # def __setslice__(self, i, j, other):
    #     list.__setslice__(self, i, j, other)
    #     self.changed()
    #
    # def __delslice__(self, i, j):
    #     list.__delslice__(self, i, j)
    #     self.changed()

    def __iadd__(self, other):
        result = list.__iadd__(self, other)
        self.changed()
        return result

    def __imul__(self, n):
        result = list.__imul__(self, n)
        self.changed()
        return result

    def append(self, item: Any) -> None:
        list.append(self, item)
        self.changed()

    def pop(self, i=-1):
        item = list.pop(self, i)
        self.changed()
        return item

    def remove(self, item) -> None:
        list.remove(self, item)
        self.changed()

    def reverse(self) -> None:
        list.reverse(self)
        self.changed()

    def sort(self, *args, **kwargs) -> None:
        list.sort(self, *args, **kwargs)
        self.changed()

    def extend(self, other) -> None:
        list.extend(self, other)
        self.changed()


class JSON(sa.types.TypeDecorator):
    """Stores any structure serializable with json.

    Usage JSON() Takes same parameters as sqlalchemy.types.Text
    """

    impl = sa.types.Text

    def process_bind_param(self, value: Any, dialect: Dialect) -> str | None:
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(
        self, value: str | None, dialect: Dialect
    ) -> dict[str, Any] | list[int] | None:
        if value is not None:
            value = json.loads(value)
        return value


class JSONUniqueListType(JSON):
    """Store a list in JSON format, with items made unique and sorted."""

    @property
    def python_type(self):
        return MutationList

    def process_bind_param(self, value, dialect):
        # value may be a simple string used in a LIKE clause for instance, so we
        # must ensure we uniquify/sort only for list-like values
        if value is not None and isinstance(value, (tuple, list)):
            value = sorted(set(value))

        return super().process_bind_param(value, dialect)


def JSONDict(*args, **kwargs):
    """Stores a dict as JSON on database, with mutability support."""
    return MutationDict.as_mutable(JSON(*args, **kwargs))


def JSONList(*args, **kwargs):
    """Stores a list as JSON on database, with mutability support.

    If kwargs has a param `unique_sorted` (which evaluated to True),
    list values are made unique and sorted.
    """
    type_ = JSON
    with contextlib.suppress(KeyError):
        if kwargs.pop("unique_sorted"):
            type_ = JSONUniqueListType

    return MutationList.as_mutable(type_(*args, **kwargs))


# TODO: replace byt UUIDType from sqlalchemy_util
class UUID(sa.types.TypeDecorator):
    """Platform-independent UUID type.

    Uses Postgresql's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    From SQLAlchemy documentation.
    """

    impl = sa.types.CHAR

    def load_dialect_impl(self, dialect: Dialect) -> CHAR:
        if dialect.name == "postgresql":
            return dialect.type_descriptor(sa.dialects.postgresql.UUID())
        return dialect.type_descriptor(sa.types.CHAR(32))

    def process_bind_param(
        self, value: None | str | uuid.UUID, dialect: Dialect
    ) -> str | None:
        if value is None:
            return value
        if dialect.name == "postgresql":
            return str(value)
        if not isinstance(value, uuid.UUID):
            value = uuid.UUID(value)
        # hexstring
        return value.hex

    def process_result_value(
        self, value: str | None, dialect: Dialect
    ) -> uuid.UUID | None:
        return value if value is None else uuid.UUID(value)

    def compare_against_backend(self, dialect, conn_type):
        if dialect.name == "postgresql":
            return isinstance(conn_type, sa.dialects.postgresql.UUID)
        return isinstance(conn_type, sa.types.CHAR)


class Locale(sa.types.TypeDecorator):
    """Store a :class:`babel.Locale` instance."""

    impl = sa.types.UnicodeText

    @property
    def python_type(self):
        return babel.Locale

    def process_bind_param(self, value: Any | None, dialect: Dialect) -> Any | None:
        if value is None:
            return None

        if not isinstance(value, babel.Locale):
            if not isinstance(value, str):
                msg = f"Unknown locale value: {format(repr(value))}"
                raise TypeError(msg)
            if not value.strip():
                return None
            value = babel.Locale.parse(value)

        code = str(value.language)
        if value.territory or value.script:
            code += f"_{value.territory!s}"

        return code

    def process_result_value(self, value: Any | None, dialect: Dialect) -> Any | None:
        return None if value is None else babel.Locale.parse(value)


class Timezone(sa.types.TypeDecorator):
    """Store a :class:`pytz.tzfile.DstTzInfo` instance."""

    impl = sa.types.UnicodeText

    @property
    def python_type(self):
        return pytz.tzinfo.DstTzInfo

    def process_bind_param(self, value: Any | None, dialect: Dialect) -> Any | None:
        if value is None:
            return None

        if not isinstance(value, pytz.tzinfo.DstTzInfo):
            if not isinstance(value, str):
                msg = f"Unknown timezone value: {value!r}"
                raise TypeError(msg)
            if not value.strip():
                return None
            value = babel.dates.get_timezone(value)

        return value.zone

    def process_result_value(self, value: Any | None, dialect: Dialect) -> Any | None:
        return None if value is None else babel.dates.get_timezone(value)



# Copyright (c) 2012-2024, Abilian SAS

""""""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from flask import Blueprint

from abilian.services.security import ANONYMOUS, Role

if TYPE_CHECKING:
    from collections.abc import Callable, Collection

    from abilian.core.models.subjects import User
    from abilian.services.security.service import SecurityService


def allow_anonymous(user: User, roles: Collection[Role], **kwargs: Any) -> bool:
    return True


def allow_access_for_roles(roles: Collection[Role] | Role) -> Callable:
    """Access control helper to check user's roles against a list of valid
    roles."""
    if isinstance(roles, Role):
        roles = (roles,)
    valid_roles = frozenset(roles)

    if ANONYMOUS in valid_roles:
        return allow_anonymous

    # FIXME: parameter not used. Why?
    def check_role(user: User, roles: Collection[Role], **kwargs):
        from abilian.services import get_service

        security: SecurityService = cast("SecurityService", get_service("security"))
        return security.has_role(user, valid_roles)

    return check_role


class AccessControlBlueprint(Blueprint):
    """An enhanced :class:`flask.blueprints.Blueprint` with access control
    helpers."""

    def __init__(
        self,
        name: str,
        import_name: str,
        allowed_roles: None | str | Role | Collection[Role] = None,
        **kwargs: Any,
    ) -> None:
        """
        :param allowed_roles: role or list of roles required to access any view in this
            blueprint.
        """
        super().__init__(name, import_name, **kwargs)

        if allowed_roles is not None:
            if isinstance(allowed_roles, str):
                allowed_roles = Role(allowed_roles)

            if isinstance(allowed_roles, Role):
                allowed_roles = (allowed_roles,)
        else:
            allowed_roles = ()

        if allowed_roles:
            self.record_once(
                lambda s: s.app.add_access_controller(
                    self.name, allow_access_for_roles(allowed_roles)
                )
            )

    def allow_any(self, func) -> None:
        pass



# Copyright (c) 2012-2024, Abilian SAS

""""""

from __future__ import annotations

from typing import TYPE_CHECKING

from .extension import CommentExtension

if TYPE_CHECKING:
    from flask import Flask


def register_plugin(app: Flask) -> None:
    CommentExtension(app)



#!/usr/bin/env python
from __future__ import annotations

import os


def sh(cmd):
    print(cmd)
    assert os.system(cmd) == 0
    print()


# sh("flask config")
# sh("flask healthcheck")
sh("gunicorn --bind 0.0.0.0:8080 -w 2 'app.flask.main:create_app()'")



from __future__ import annotations

import nox

PYTHON_VERSIONS = ["3.10", "3.11", "3.12", "3.13"]
DB_URIS = [
    "sqlite:///test.db",
    # TODO: Postgres & MariaDB
]
nox.options.reuse_existing_virtualenvs = True


@nox.session
def lint(session: nox.Session) -> None:
    session.run("uv", "sync")
    session.run("uv", "pip", "check")
    session.run("uv", "run", "make", "lint")
    session.run("uv", "pip", "install", "safety", "pip-audit")
    session.run("uv", "run", "adt", "audit")


@nox.session(python=PYTHON_VERSIONS)
def pytest(session: nox.Session) -> None:
    session.run("uv", "sync")
    session.run("uv", "run", "pytest", "--tb=short")


@nox.session
@nox.parametrize("db_uri", DB_URIS)
def db_test(session: nox.Session, db_uri: str) -> None:
    env = {
        "FLASK_SQLALCHEMY_DATABASE_URI": db_uri,
    }
    session.run("uv", "sync")
    session.run("uv", "run", "pytest", "--tb=short", env=env)



# Copyright (c) 2012-2024, Abilian SAS

"""Static configuration for the application.

TODO: add more (runtime) flexibility in plugin discovery, selection
and activation.
"""

from __future__ import annotations

import jinja2

from abilian.app import Application as BaseApplication, create_app as base_create_app
from abilian.core.dramatiq.setup import init_dramatiq_engine
from abilian.core.plugin_manager import CORE_PLUGINS
from abilian.services import converter

from .apps.documents.repository import content_repository
from .extension import sbe

__all__ = ["create_app"]

SBE_PLUGINS = [
    *CORE_PLUGINS,
    "abilian.sbe.apps.main",
    "abilian.sbe.apps.notifications",
    "abilian.sbe.apps.preferences",
    "abilian.sbe.apps.wiki",
    "abilian.sbe.apps.wall",
    "abilian.sbe.apps.documents",
    "abilian.sbe.apps.forum",
    # "abilian.sbe.apps.calendar",
    "abilian.sbe.apps.communities",
    "abilian.sbe.apps.social",
    "abilian.sbe.apps.preferences",
]


def create_app(config: type | None = None, **kw) -> BaseApplication:
    kw["plugins"] = SBE_PLUGINS
    app = base_create_app(config=config, **kw)

    with app.app_context():
        content_repository.init_app(app)
        converter.init_app(app)

        loader = jinja2.PackageLoader("abilian.sbe")
        app.register_jinja_loaders(loader)

    sbe.init_app(app)
    init_dramatiq_engine(app)

    return app



# Copyright (c) 2012-2024, Abilian SAS

from __future__ import annotations

from flask import Blueprint

notifications = Blueprint(
    "notifications",
    __name__,
    url_prefix="/notifications",
    template_folder="../templates",
)



# Copyright (c) 2012-2024, Abilian SAS

"""Mail extension for Abilian Core."""

from __future__ import annotations

from typing import Any

import flask_mail
from flask import current_app

from abilian.core.logger_patch import patch_logger


# patch flask.ext.mail.Message.send to always set enveloppe_from default mail
# sender
# FIXME: we'ld rather subclass Message and update all imports
def _message_send(self: Any, connection: flask_mail.Connection) -> None:
    """Send a single message instance.

    If TESTING is True the message will not actually be sent.

    :param self: a Message instance.
    """
    sender = current_app.config["MAIL_SENDER"]
    if not self.extra_headers:
        self.extra_headers = {}
    self.extra_headers["Sender"] = sender
    connection.send(self, sender)


patch_logger.info(flask_mail.Message.send)
flask_mail.Message.send = _message_send

mail = flask_mail.Mail()



# Copyright (c) 2012-2024, Abilian SAS

"""Register extensions as a plugin.

NOTE: panels are currently loaded and registered manually. This may change
in the future.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from flask import Flask


def register_plugin(app: Flask) -> None:
    from .main import blueprint

    app.register_blueprint(blueprint)



# Copyright (c) 2012-2024, Abilian SAS

"""Additional data types for sqlalchemy."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from flask_sqlalchemy import SQLAlchemy as SAExtension

if TYPE_CHECKING:
    from flask import Flask
    from sqlalchemy.engine.url import URL


class SQLAlchemy(SAExtension):
    """Base subclass of :class:`flask_sqlalchemy.SQLAlchemy`.

    Add our custom driver hacks.
    """

    def apply_driver_hacks(
        self, app: Flask, info: URL, options: dict[str, Any]
    ) -> None:
        super().apply_driver_hacks(app, info, options)

        if info.drivername == "sqlite":
            connect_args = options.setdefault("connect_args", {})

            if "isolation_level" not in connect_args:
                # required to support savepoints/rollback without error. It disables
                # implicit BEGIN/COMMIT statements made by pysqlite (a COMMIT kills all
                # savepoints made).
                connect_args["isolation_level"] = None

        elif info.drivername.startswith("postgres"):
            options.setdefault("client_encoding", "utf8")
